#  RCAIDE/Methods/Energy/Distributors/Electrical_Bus/append_battery_conditions.py
# 
# Created: Sep 2024, S. Shekar

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE imports  
import RCAIDE
from RCAIDE.Framework.Mission.Common     import   Conditions
# ----------------------------------------------------------------------------------------------------------------------
#  METHODS
# ---------------------------------------------------------------------------------------------------------------------- 
def append_bus_conditions(bus,segment): 
    """
    Appends conditions for the electrical bus to the segment's energy conditions dictionary.

    Parameters
    ----------
    bus : RCAIDE.Library.Components.Distributors.ElectricalBus
    
    Returns
    -------
    None
        This function modifies the segment.state.conditions.energy dictionary in-place.
    
    Notes
    -----
    This function creates a Conditions object for the electrical bus within the segment's
    energy conditions dictionary, indexed by the bus tag. It initializes various bus
    properties as zero arrays with the same length as the segment's state vector.
    
    The initialized properties include:
        - Battery module conditions
        - Fuel cell stack conditions
        - Power draw
        - State of charge and depth of discharge
        - Current draw and charging current
        - Voltage (open circuit and under load)
        - Heat energy generated
        - Efficiency
        - Temperature
        - Energy
        - Regenerative power
    
    For segments with an initial battery state of charge specified, the function also
    sets the initial energy and state of charge values accordingly.
    
    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Distributors.Electrical_Bus.compute_bus_conditions
    """
    ones_row                                                                     = segment.state.ones_row
               
    segment.state.conditions.energy[bus.tag]                                     = Conditions()
    segment.state.conditions.energy[bus.tag].battery_modules                     = Conditions()
    segment.state.conditions.energy[bus.tag].fuel_cell_stacks                    = Conditions()
    segment.state.conditions.energy[bus.tag].power_draw                          = 0 * ones_row(1)
    segment.state.conditions.energy[bus.tag].hybrid_power_split_ratio            = segment.hybrid_power_split_ratio * ones_row(1)
    segment.state.conditions.energy[bus.tag].battery_fuel_cell_power_split_ratio = segment.battery_fuel_cell_power_split_ratio * ones_row(1) 
    segment.state.conditions.energy[bus.tag].state_of_charge                     = 0 * ones_row(1) 
    segment.state.conditions.energy[bus.tag].depth_of_discharge                  = 0 * ones_row(1) 
    segment.state.conditions.energy[bus.tag].current_draw                        = 0 * ones_row(1)
    segment.state.conditions.energy[bus.tag].charging_current                    = 0 * ones_row(1)
    segment.state.conditions.energy[bus.tag].voltage_open_circuit                = 0 * ones_row(1)
    segment.state.conditions.energy[bus.tag].voltage_under_load                  = 0 * ones_row(1) 
    segment.state.conditions.energy[bus.tag].heat_energy_generated               = 0 * ones_row(1) 
    segment.state.conditions.energy[bus.tag].efficiency                          = 0 * ones_row(1)
    segment.state.conditions.energy[bus.tag].temperature                         = 0 * ones_row(1)
    segment.state.conditions.energy[bus.tag].energy                              = 0 * ones_row(1)
    segment.state.conditions.energy[bus.tag].regenerative_power                  = 0 * ones_row(1)

     # first segment  
    if 'initial_battery_state_of_charge' in segment:  
        initial_battery_energy                                             = segment.initial_battery_state_of_charge*bus.maximum_energy   
        segment.state.conditions.energy[bus.tag].maximum_initial_energy    = initial_battery_energy
        segment.state.conditions.energy[bus.tag].energy                    = initial_battery_energy* ones_row(1)
        segment.state.conditions.energy[bus.tag].state_of_charge           = segment.initial_battery_state_of_charge* ones_row(1) 
        segment.state.conditions.energy[bus.tag].depth_of_discharge        = 1 - segment.initial_battery_state_of_charge* ones_row(1)
   
    return


def append_bus_segment_conditions(bus,conditions,segment):
    """
    Sets the initial bus properties at the start of each segment based on the last point from the previous segment.
    
    Parameters
    ----------
    bus : ElectricalBus
        The electrical bus component for which conditions are being initialized.
    conditions : dict
        Dictionary containing conditions from the previous segment.
    segment : Segment
        The current mission segment in which the bus is operating.
    
    Returns
    -------
    None
    
    Notes
    -----
    This function initializes the power draw for the electrical bus at the start of a new segment
    by transferring thermal power draw information from the previous segment. It handles power
    requirements from battery thermal management systems (BTMS) and heat exchangers.
    
    For segments with initial conditions from a previous segment, the function also:
        1. Sets the battery discharge flag based on the segment type (false for recharge segments)
        2. Transfers the final energy state from the previous segment to the initial state of the current segment
    
    This ensures continuity of energy states between mission segments.
    
    **Major Assumptions**
        * Battery recharge segments are identified by their class type
        * Power draw is initialized from thermal management components
    
    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Distributors.Electrical_Bus.append_bus_conditions
    RCAIDE.Library.Methods.Powertrain.Distributors.Electrical_Bus.compute_bus_conditions
    """    
    bus_conditions             = conditions[bus.tag]
    ones_row                   = segment.state.ones_row
    bus_conditions.power_draw  = 0 * ones_row(1)
    # Thermal power draw
    if segment.state.initials:
        for network in segment.analyses.energy.vehicle.networks:
            for coolant_line in  network.coolant_lines:
                for tag, item in  coolant_line.items():
                    if tag == 'battery_modules':
                        for battery in item:
                            for btms in  battery:
                                bus_conditions.power_draw[0,0]   +=  segment.state.initials.conditions.energy[coolant_line.tag][btms.tag].power[-1] 
                    if tag == 'heat_exchangers':
                        for heat_exchanger in  item:                    
                            bus_conditions.power_draw[0,0]   +=  segment.state.initials.conditions.energy[coolant_line.tag][heat_exchanger.tag].power[-1] 
        # Bus Properties 
        bus_initials            = segment.state.initials.conditions.energy[bus.tag]
        if type(segment) ==  RCAIDE.Framework.Mission.Segments.Ground.Battery_Recharge:             
            bus_initials.battery_discharge_flag           = False 
        else:                   
            bus_initials.battery_discharge_flag           = True     
        bus_conditions.energy[0,0]          = bus_initials.energy[-1,0]


    return


# RCAIDE/Library/Attributes/Propellants/__init__.py
# 

"""
This module provides propellant definitions for RCAIDE simulations. It includes classes 
for various propellant types categorized as follows:

**Aviation Fuels**
    - Aviation_Gasoline : Aviation gasoline properties
    - Jet_A : Commercial aviation kerosene
    - Jet_A1 : International spec aviation kerosene
    - JP7 : Advanced high thermal stability jet fuel

**Rocket Propellants**
    - Rocket_LH2 : Rocket-grade liquid hydrogen
    - Rocket_RP1 : Rocket-grade kerosene

**Cryogenic Fuels**
    - Liquid_Hydrogen : LH2 for aviation applications
    - Liquid_Natural_Gas : LNG fuel properties
    - Gaseous_Hydrogen : GH2 properties

**Hydrocarbon Fuels**
    - Methane : CH4 properties
    - Propane : C3H8 properties
    - Ethane : C2H6 properties
    - Alkane_Mixture : Customizable alkane blend properties

**Alcohols**
    - Ethanol : C2H5OH properties
    - Butanol : C4H9OH properties
    - Propanol : C3H7OH properties
    - Alcohol_Mixture : Customizable alcohol blend properties

**Base Classes**
    - Propellant : Base class defining standard propellant interfaces

See Also
--------
RCAIDE.Library.Methods.Propulsors : Propulsion analysis tools
RCAIDE.Library.Components.Propulsors : Propulsion system components
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .Propellant           import Propellant
from .Aviation_Gasoline    import Aviation_Gasoline
from .Jet_A                import Jet_A
from .Jet_A1               import Jet_A1 
from .Gaseous_Hydrogen     import Gaseous_Hydrogen
from .Liquid_Natural_Gas   import Liquid_Natural_Gas
from .Liquid_Petroleum_Gas import Liquid_Petroleum_Gas
from .JP7                  import JP7
from .Methane              import Methane
from .Propane              import Propane
from .Ethane               import Ethane
from .Ethanol              import Ethanol
from .Liquid_Hydrogen      import Liquid_Hydrogen
from .Rocket_LH2           import Rocket_LH2
from .Rocket_RP1           import Rocket_RP1
from .Butanol		   import Butanol
from .Propanol		   import Propanol
from .Alkane_Mixture	   import Alkane_Mixture
from .Alcohol_Mixture	   import Alcohol_Mixture


# RCAIDE/Library/Attributes/Propellants.py
# 
# 
# Created:  Sep 2023, M. Clarke
# Modified: 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 

from RCAIDE.Framework.Core import Data
from RCAIDE.Library.Components.Mass_Properties import Mass_Properties

# ----------------------------------------------------------------------------------------------------------------------
#  Propellant
# ----------------------------------------------------------------------------------------------------------------------  
class Propellant(Data):
    """
    Base class for defining propellant properties and characteristics in propulsion systems.

    Attributes
    ----------
    tag : str
        Identifier for the specific propellant type
    reactant : str
        Primary oxidizer used for combustion
    density : float
        Fuel density in kg/m³
    specific_energy : float
        Specific energy content in J/kg
    energy_density : float
        Energy density in J/m³
    lower_heating_value : float
        Lower heating value in J/kg
    mass_properties : Mass_Properties
        Object containing mass-related properties
    max_mass_fraction : Data
        Maximum fuel-to-oxidizer mass ratios
            - Air : float
                Maximum mass fraction with air
            - O2 : float
                Maximum mass fraction with pure oxygen
    temperatures : Data
        Critical temperatures in K
            - flash : float
                Flash point temperature
            - autoignition : float
                Autoignition temperature
            - freeze : float
                Freezing point temperature
            - boiling : float
                Boiling point temperature
    emission_indices : Data
        Emission indices in kg/kg fuel
            - Production : float
                CO2 production rate
            - CO2 : float
                Carbon dioxide emissions
            - H2O : float
                Water vapor emissions
            - SO2 : float
                Sulfur dioxide emissions
            - NOx : float
                Nitrogen oxides emissions
            - Soot : float
                Particulate matter emissions
    global_warming_potential_100 : Data
        100-year global warming potentials
            - CO2 : float
                Carbon dioxide impact
            - H2O : float
                Water vapor impact
            - SO2 : float
                Sulfur dioxide impact
            - NOx : float
                Nitrogen oxides impact
            - Soot : float
                Particulate matter impact
            - Contrails : float
                Contrail formation impact

    Notes
    -----
    This base class provides a standardized structure for defining propellant 
    properties, including physical characteristics, combustion parameters, and 
    environmental impacts. It serves as a template for specific propellant 
    implementations.

    **Definitions**
    
    'Emission Index'
        Mass of pollutant produced per unit mass of fuel burned
    
    'Global Warming Potential'
        Relative measure of heat trapped in atmosphere compared to CO2
    
    'Lower Heating Value'
        Heat of combustion excluding latent heat of water vapor
    """

    def __defaults__(self):
        """This sets the default values. 
    
        Assumptions:
            None
        
        Source:
            None
        """    
        self.tag                       = 'Propellant'
        self.reactant                  = 'O2'
        self.density                   = 0.0                       # kg/m^3
        self.specific_energy           = 0.0                       # MJ/kg
        self.energy_density            = 0.0                       # MJ/m^3
        self.lower_heating_value       = 0.0                       # MJ/kg
        self.mass_properties           = Mass_Properties()
        self.max_mass_fraction         = Data({'Air' : 0.0, 'O2' : 0.0}) # kg propellant / kg oxidizer
        self.temperatures              = Data()
        self.temperatures.flash        = 0.0                       # K
        self.temperatures.autoignition = 0.0                       # K
        self.temperatures.freeze       = 0.0                       # K
        self.temperatures.boiling      = 0.0                       # K

        self.stoichiometric_fuel_air_ratio = 0         # [-] Stoichiometric Fuel to Air ratio
        self.heat_of_vaporization          = 0         # [J/kg] Heat of vaporization at standard conditions
        self.temperature                   = 0         # [K] Temperature of fuel
        self.pressure                      = 0         # [Pa] Pressure of fuel
        self.fuel_surrogate_S1             = {} # [-] Mole fractions of fuel surrogate species
        self.kinetic_mechanism             = '' # [-] Kinetic mechanism for fuel surrogate species
        self.oxidizer                      = ''
        
        # Emission Indices 
        self.emission_indices             =  Data() 
        self.emission_indices.Production  = 0
        self.emission_indices.CO2         = 0
        self.emission_indices.H2O         = 0
        self.emission_indices.SO2         = 0
        self.emission_indices.NOx         = 0
        self.emission_indices.Soot        = 0 

        self.global_warming_potential_100           =  Data() 
        self.global_warming_potential_100.CO2       = 0
        self.global_warming_potential_100.H2O       = 0
        self.global_warming_potential_100.SO2       = 0
        self.global_warming_potential_100.NOx       = 0
        self.global_warming_potential_100.Soot      = 0  
        self.global_warming_potential_100.Contrails = 0        


# RCAIDE/Methods/Powertrain/Sources/Batteries/Ragone/find_ragone_properties.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE imports 
from .size_module_from_energy_and_power  import size_module_from_energy_and_power
from .find_specific_power                import find_specific_power

# ----------------------------------------------------------------------------------------------------------------------
#  METHOD
# ----------------------------------------------------------------------------------------------------------------------  
def find_ragone_properties(specific_energy, battery, energy, power):
    """
    Determines battery mass based on specific energy, energy required, and power required.
    
    Parameters
    ----------
    specific_energy : float
        Specific energy value to use for the battery [J/kg]
    battery : Battery
        The battery component to be sized
    energy : float
        Required energy capacity [J]
    power : float
        Required power output [W]
    
    Returns
    -------
    mass : float
        Calculated mass of the battery [kg]
    
    Notes
    -----
    This function calculates the mass of a battery needed to meet both energy and power
    requirements, using a specific energy value and the corresponding specific power
    determined from a Ragone curve.
    
    The function performs the following steps:
      1. Calculates the specific power corresponding to the given specific energy using the Ragone curve correlation
      2. Sizes the battery module based on the energy and power requirements
    
    The Ragone curve describes the trade-off between specific energy and specific power
    in energy storage devices. By using this relationship, the function ensures that
    the battery is sized appropriately to meet both energy and power requirements.
    
    **Major Assumptions**
        * The Ragone curve accurately represents the energy-power trade-off
        * Linear scaling of energy and power with mass
    
    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Sources.Batteries.Common.find_specific_power
    RCAIDE.Library.Methods.Powertrain.Sources.Batteries.Common.size_module_from_energy_and_power
    """
    
    find_specific_power(battery, specific_energy)
    size_module_from_energy_and_power(battery, energy, power)
    
    # can be used for a simple optimization
    return battery.mass_properties.mass 


# RCAIDE/Library/Methods/Powertrain/Propulsors/Electric_Rotor_Propulsor/append_electric_rotor_conditions.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke  

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE imports
import RCAIDE
from RCAIDE.Framework.Mission.Common                      import Conditions

# ---------------------------------------------------------------------------------------------------------------------- 
#  append electric rotor network conditions
# ----------------------------------------------------------------------------------------------------------------------    
def append_electric_rotor_conditions(propulsor, segment, energy_conditions, noise_conditions):
    """
    Appends data structures arrays for storing electric rotor conditions.
    
    Parameters
    ----------
    propulsor : RCAIDE.Library.Components.Propulsors.Electric_Rotor
        Electric rotor propulsor component with the following attributes:
            - tag : str
                Identifier for the propulsor
            - items : dict
                Dictionary of subcomponents
    segment : RCAIDE.Framework.Mission.Segments.Segment
        Mission segment with the following attributes:
            - state : Data
                Segment state
                - ones_row : function
                    Function to create array of ones with specified length
    energy_conditions : RCAIDE.Framework.Mission.Common.Conditions
        Energy conditions container where electric rotor conditions will be stored
    noise_conditions : RCAIDE.Framework.Mission.Common.Conditions
        Noise conditions container where electric rotor noise conditions will be stored
    
    Returns
    -------
    None
        Results are stored in energy_conditions.propulsors[propulsor.tag] and
        noise_conditions.propulsors[propulsor.tag]
    
    Notes
    -----
    This function initializes the necessary data structures for storing electric rotor
    operating conditions during a mission segment. It creates zero-filled arrays
    for various performance parameters and recursively calls the append_operating_conditions
    method for each subcomponent of the electric rotor propulsor.
    
    The function initializes the following parameters in energy_conditions:
        - throttle
        - commanded_thrust_vector_angle
        - thrust
        - power
        - moment
    
    It also creates a noise conditions container for the electric rotor.
    
    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Propulsors.Electric_Rotor.compute_electric_rotor_performance
    """
    # unpack 
    ones_row          = segment.state.ones_row 
    
    # add propulsor conditions              
    energy_conditions.propulsors[propulsor.tag]                               = Conditions()  
    energy_conditions.propulsors[propulsor.tag].throttle                      = 0. * ones_row(1)      
    energy_conditions.propulsors[propulsor.tag].commanded_thrust_vector_angle = 0. * ones_row(1)  
    energy_conditions.propulsors[propulsor.tag].thrust                        = 0. * ones_row(3) 
    energy_conditions.propulsors[propulsor.tag].power                         = 0. * ones_row(1) 
    energy_conditions.propulsors[propulsor.tag].moment                        = 0. * ones_row(3)  
    noise_conditions.propulsors[propulsor.tag]                                = Conditions() 
       
    # parse propulsor for comoonent and append 
    for tag, item in  propulsor.items(): 
        if issubclass(type(item), RCAIDE.Library.Components.Component):
            item.append_operating_conditions(segment,energy_conditions,noise_conditions)
    return



# RCAIDE/Library/Attributes/Propellants/Jet_A.py
# 
#
# Created:  Mar 2024, M. Clarke

# ---------------------------------------------------------------------------------------------------------------------- 
#  Imports
# ---------------------------------------------------------------------------------------------------------------------- 

from .Propellant import Propellant
from RCAIDE.Framework.Core import Data  
# ---------------------------------------------------------------------------------------------------------------------- 
#  Jet_A1 Propellant Class
# ----------------------------------------------------------------------------------------------------------------------  
class Jet_A(Propellant):
    """
    A class representing Jet A aviation kerosene fuel properties and emissions characteristics.

    Attributes
    ----------
    tag : str
        Identifier for the propellant ('Jet_A')
    reactant : str
        Oxidizer used for combustion ('O2')
    density : float
        Fuel density in kg/m³ (820.0)
    specific_energy : float
        Specific energy content in J/kg (43.02e6)
    energy_density : float
        Energy density in J/m³ (35276.4e6)
    lower_heating_value : float
        Lower heating value in J/kg (43.24e6)
    max_mass_fraction : Data
        Maximum fuel-to-oxidizer mass ratios
            - Air : float
                Maximum mass fraction with air (0.0633)
            - O2 : float
            Maximum mass fraction with pure oxygen (0.3022)
    use_high_fidelity_kinetics_model : bool
        Flag for using detailed chemical kinetics (True)
    fuel_surrogate_chemical_properties : dict
        Simplified surrogate composition
            - N-C12H26 : float
                Dodecane fraction (0.6)
            - A1CH3 : float
                Toluene fraction (0.2)
            - A1 : float
                Benzene fraction (0.2)
    fuel_chemical_properties : dict
        Detailed chemical composition for high-fidelity model
    emission_indices : Data
        Emission indices in kg/kg fuel
            - Production : float
                CO2 production rate (0.4656)
            - CO2 : float
                Carbon dioxide (3.16)
            - CO : float
                Carbon monoxide (0.00201)
            - H2O : float
                Water vapor (1.34)
            - SO2 : float
                Sulfur dioxide (0.0012)
            - NOx : float
                Nitrogen oxides (0.01514)
            - Soot : float
                Particulate matter (0.0012)
    global_warming_potential_100 : Data
        100-year global warming potentials
            - CO2 : float
                Carbon dioxide (1)
            - H2O : float
                Water vapor (0.06)
            - SO2 : float
                Sulfur dioxide (-226)
            - NOx : float
                Nitrogen oxides (52)
            - Soot : float
                Particulate matter (1166)
            - Contrails : float
                Contrail formation (11)

    Notes
    -----
    This class implements properties for Jet A aviation kerosene, including both 
    simplified and detailed chemical kinetics options. Properties are specified at 
    standard conditions (15°C, 1 atm).

    **Definitions**
    
    'Emission Index'
        Mass of pollutant produced per mass of fuel burned
    
    'Global Warming Potential'
        Relative measure of heat trapped in atmosphere compared to CO2
    
    'Surrogate Model'
        Simplified three-component representation of complex fuel mixture

    **Major Assumptions**
        * Properties are for standard temperature and pressure conditions (15C, 1atm)
        * Surrogate model uses three-component representation
        * Detailed model includes complex hydrocarbon mixture
        * Emission indices are for typical aircraft cruise conditions

    References
    ----------
    [1] D.S. Lee, D.W. Fahey, A. Skowron, M.R. Allen, U. Burkhardt, Q. Chen, S.J. Doherty, S. Freeman, P.M. Forster, J. Fuglestvedt, A. Gettelman, R.R. De León, L.L. Lim, M.T. Lund, R.J. Millar, B. Owen, J.E. Penner, G. Pitari, M.J. Prather, R. Sausen, L.J. Wilcox, "The contribution of global aviation to anthropogenic climate forcing for 2000 to 2018," Atmospheric Environment, Volume 244, 2021, 117834, ISSN 1352-2310, https://doi.org/10.1016/j.atmosenv.2020.117834.
    [2] Chevron Products Company. (n.d.). Aviation fuels. Chevron. https://www.chevron.com/-/media/chevron/operations/documents/aviation-tech-review.pdf 
    """

    def __defaults__(self):
        """This sets the default values.
    
        Assumptions:
            None
        
        Source:
            Emission Indices from Lee, David S., et al. "The contribution of global aviation to anthropogenic climate forcing
            for 2000 to 2018." Atmospheric environment 244 (2021): 117834
            
            
            
            
        """    
        self.tag                           = 'Jet_A'
        self.reactant                      = 'O2'
        self.density                       = 820.0                          # kg/m^3 (15 C, 1 atm)
        self.specific_energy               = 43.02e6                        # J/kg
        self.energy_density                = 35276.4e6                      # J/m^3
        self.lower_heating_value           = 43.24e6                        # J/kg 
        self.heat_of_vaporization          = 300000                         # J/kg
        self.stoichiometric_fuel_air_ratio = 0.068
        self.max_mass_fraction             = Data({'Air' : 0.0633,'O2' : 0.3022})   # kg propellant / kg oxidizer
   
        self.stoichiometric_fuel_air_ratio = 0         # [-] Stoichiometric Fuel to Air ratio
        self.heat_of_vaporization          = 0         # [J/kg] Heat of vaporization at standard conditions
        self.temperature                   = 0         # [K] Temperature of fuel
        self.pressure                      = 0         # [Pa] Pressure of fuel
        self.fuel_surrogate_S1             = {} # [-] Mole fractions of fuel surrogate species
        self.kinetic_mechanism             = '' # [-] Kinetic mechanism for fuel surrogate species
        self.oxidizer                      = ''

        # critical temperatures   
        self.temperatures.flash           = 311.15                 # K
        self.temperatures.autoignition    = 483.15                 # K
        self.temperatures.freeze          = 233.15                 # K
        self.temperatures.boiling         = 0.0                    # K 

        self.emission_indices.Production  = 0.4656   # kg/kg Greet 
        self.emission_indices.CO2         = 3.16    # kg/kg
        self.emission_indices.CO          = 0.00201 # kg/kg
        self.emission_indices.H2O         = 1.23    # kg/kg  
        self.emission_indices.SO2         = 0.0012  # kg/kg
        self.emission_indices.NOx         = 0.01514 # kg/kg
        self.emission_indices.Soot        = 0.0012  # kg/kg
        
        self.global_warming_potential_100.CO2       = 1     # CO2e/kg  
        self.global_warming_potential_100.H2O       = 0.06  # CO2e/kg  
        self.global_warming_potential_100.SO2       = -226  # CO2e/kg  
        self.global_warming_potential_100.NOx       = 52    # CO2e/kg  
        self.global_warming_potential_100.Soot      = 1166  # CO2e/kg    
        self.global_warming_potential_100.Contrails = 11 # kg/CO2e/km  


# RCAIDE/Library/Methods/Powertrain/Propulsors/Internal_Combustion_Engine/compute_internal_combustion_engine_performance.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE imports  
from RCAIDE.Framework.Core import Units  
from RCAIDE.Library.Methods.Powertrain.Converters.Engine import compute_power_from_throttle
from RCAIDE.Library.Methods.Powertrain.Converters.Rotor.compute_rotor_performance import  compute_rotor_performance

# pacakge imports  
from copy import deepcopy
import numpy as np  

# ----------------------------------------------------------------------------------------------------------------------
# compute_internal_combustion_engine_performance
# ---------------------------------------------------------------------------------------------------------------------- 
def compute_internal_combustion_engine_performance(propulsor, state, center_of_gravity=[[0.0, 0.0, 0.0]]):
    """
    Computes the performance of an internal combustion engine propulsion system.
    
    Parameters
    ----------
    propulsor : RCAIDE.Library.Components.Propulsors.Internal_Combustion_Engine
        Internal combustion engine propulsor component with the following attributes:
            - tag : str
                Identifier for the propulsor
            - engine : Data
                Engine component
                    - tag : str
                        Identifier for the engine
            - propeller : Data
                Propeller component
                    - tag : str
                        Identifier for the propeller
                    - origin : list of lists
                        Origin coordinates [[x, y, z]] [m]
    state : RCAIDE.Framework.Mission.Common.State
        State object containing:
            - conditions : Data
                Flight conditions
                    - energy : Data
                        Energy conditions
                            - propulsors : dict
                                Propulsor energy conditions indexed by tag
                            - converters : dict
                                Converter energy conditions indexed by tag
            - ones_row : function
                Function to create array of ones with specified length
    center_of_gravity : list of lists, optional
        Center of gravity coordinates [[x, y, z]] [m]
        Default: [[0.0, 0.0, 0.0]]
    
    Returns
    -------
    thrust : numpy.ndarray
        Thrust force vector [N]
    moment : numpy.ndarray
        Moment vector [N·m]
    power : numpy.ndarray
        Mechanical power [W]
    power_elec : numpy.ndarray
        Electrical power [W] (zero for non-hybrid engines)
    stored_results_flag : bool
        Flag indicating if results are stored
    stored_propulsor_tag : str
        Tag of the propulsor with stored results
    
    Notes
    -----
    This function computes the performance of an internal combustion engine propulsion system
    by linking the engine to the propeller. It performs the following steps:
        * Sets the engine throttle based on the propulsor throttle
        * Computes the engine power output using the throttle setting
        * Transfers the engine's angular velocity and throttle to the propeller
        * Computes the propeller performance
        * Calculates the moment generated by the propeller thrust
        * Stores the results in the conditions data structure
    
    The function assumes a direct mechanical connection between the engine and propeller,
    with no electrical power generation or consumption (non-hybrid configuration).
    
    **Major Assumptions**
        * Direct mechanical connection between engine and propeller
        * No electrical power generation or consumption
    
    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Converters.Engine.compute_power_from_throttle
    RCAIDE.Library.Methods.Powertrain.Converters.Rotor.compute_rotor_performance
    """
    conditions              = state.conditions  
    engine                  = propulsor.engine 
    propeller               = propulsor.propeller
    ice_conditions          = conditions.energy.propulsors[propulsor.tag]

    # Throttle the engine 
    conditions.energy.converters[engine.tag].throttle        = conditions.energy.propulsors[propulsor.tag].throttle 
    compute_power_from_throttle(engine,conditions)        
     
    # Run the propeller to get the power
    conditions.energy.converters[propeller.tag].omega          = conditions.energy.converters[engine.tag].omega 
    conditions.energy.converters[propeller.tag].throttle       = conditions.energy.converters[engine.tag].throttle
    conditions.energy.converters[propeller.tag].commanded_thrust_vector_angle =  conditions.energy.propulsors[propulsor.tag].commanded_thrust_vector_angle
    compute_rotor_performance(propeller,conditions)

    # Compute moment 
    moment_vector           = 0*state.ones_row(3)
    moment_vector[:,0]      = propeller.origin[0][0]  -  center_of_gravity[0][0] 
    moment_vector[:,1]      = propeller.origin[0][1]  -  center_of_gravity[0][1] 
    moment_vector[:,2]      = propeller.origin[0][2]  -  center_of_gravity[0][2]
    moment                  =  np.cross(moment_vector, conditions.energy.converters[propeller.tag].thrust)       
    
    # Create the outputs
    ice_conditions.fuel_flow_rate            = conditions.energy.converters[engine.tag].fuel_flow_rate  
    stored_results_flag                      = True
    stored_propulsor_tag                     = propulsor.tag 


    # compute total forces and moments from propulsor (future work would be to add moments from motors)
    ice_conditions.thrust      = conditions.energy.converters[propeller.tag].thrust 
    ice_conditions.moment      = moment
    ice_conditions.power       = conditions.energy.converters[propeller.tag].power  
    
    # currently, no hybridization
    power_elec =  0*state.ones_row(1)
    
    return ice_conditions.thrust,ice_conditions.moment,ice_conditions.power,power_elec,stored_results_flag,stored_propulsor_tag  
    
    
def reuse_stored_internal_combustion_engine_data(propulsor,state,network,stored_propulsor_tag,center_of_gravity= [[0.0, 0.0,0.0]]):
    '''Reuses results from one propulsor for identical propulsors
    
    Assumptions: 
    N/A

    Source:
    N/A

    Inputs:  
    conditions           - operating conditions data structure        [-]  
    fuel_line            - fuel line                                  [-] 
    propulsor            - propulsor data structure               [-] 
    total_thrust         - thrust of propulsor group              [N]
    total_power          - power of propulsor group               [W] 

    Outputs:  
    total_thrust         - thrust of propulsor group              [N]
    total_power          - power of propulsor group               [W] 
    
    Properties Used: 
    N.A.        
    ''' 

    # unpack
    conditions   = state.conditions 
    engine       = propulsor.engine
    propeller    = propulsor.propeller 
    engine_0     = network.propulsors[stored_propulsor_tag].engine
    propeller_0  = network.propulsors[stored_propulsor_tag].propeller  
    
    # deepcopy results 
    conditions.energy.propulsors[propulsor.tag]     = deepcopy(conditions.energy.propulsors[stored_propulsor_tag.tag])
    conditions.energy.converters[engine.tag]        = deepcopy(conditions.energy.converters[engine_0.tag])
    conditions.energy.converters[propeller.tag]     = deepcopy(conditions.energy.converters[propeller_0.tag])
   
    # compoment 
    thrust_vector           = conditions.energy.converters[propeller.tag].thrust 
    power                   = conditions.energy.converters[propeller.tag].power   
    moment_vector           = 0*state.ones_row(3) 
    moment_vector[:,0]      = propeller.origin[0][0]  -  center_of_gravity[0][0] 
    moment_vector[:,1]      = propeller.origin[0][1]  -  center_of_gravity[0][1] 
    moment_vector[:,2]      = propeller.origin[0][2]  -  center_of_gravity[0][2]
    moment                  =  np.cross(moment_vector, thrust_vector)
    
    # pack 
    conditions.energy.converters[propeller.tag].moment = moment  
    conditions.energy.propulsors[propulsor.tag].thrust = thrust_vector   
    conditions.energy.propulsors[propulsor.tag].moment = moment  
    conditions.energy.propulsors[propulsor.tag].power  = power
  
    power_elec =  0*state.ones_row(1)
    
    return thrust_vector,moment,power, power_elec

            
               


# RCAIDE/Methods/Performance/estimate_stall_speed.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports 
import RCAIDE
 
# Pacakge imports 
import numpy as np  

#------------------------------------------------------------------------------
# Stall Speed Estimation
#------------------------------------------------------------------------------ 
def estimate_stall_speed(vehicle_mass,reference_area,altitude,maximum_lift_coefficient): 
    """
    Calculates the stall speed of an aircraft at a given altitude and maximum lift coefficient.

    Parameters
    ----------
    vehicle_mass : float
        Total mass of the vehicle [kg]
    reference_area : float
        Wing reference area [m²]
    altitude : float
        Flight altitude [m]
    maximum_lift_coefficient : float
        Maximum lift coefficient of the aircraft [unitless]

    Returns
    -------
    V_stall : float
        Stall speed [m/s]

    Notes
    -----
    The stall speed is calculated using the standard lift equation solved for velocity:
    
    .. math::
        V_{stall} = \sqrt{\\frac{2W}{\\rho S C_{L_{max}}}}

    where:
        * W = mg (vehicle weight)
        * ρ = air density at altitude
        * S = reference area
        * CL_max = maximum lift coefficient

    **Major Assumptions**
        * Steady, level flight
        * Incompressible flow
        * Standard atmospheric conditions
        * No wind or atmospheric disturbances
        * Rigid aircraft structure

    See Also
    --------
    RCAIDE.Library.Attributes.Atmospheres
    RCAIDE.Library.Methods.Performance.estimate_take_off_field_length
    """
      
    g       = 9.81 
    atmo    = RCAIDE.Framework.Analyses.Atmospheric.US_Standard_1976()
    rho     = atmo.compute_values(altitude,0.).density 
    V_stall = float(np.sqrt((2.*vehicle_mass*g)/(rho*reference_area*maximum_lift_coefficient)))  
    
    return V_stall


# RCAIDE/Methods/Mission/Segments/Hover/__init__.py
# 

"""
Collection of mission segment types for analyzing vertical flight operations, including hover, 
vertical climb, and descent maneuvers typically associated with VTOL aircraft.

See Also
--------
RCAIDE.Library.Mission.Segments.Transition
RCAIDE.Library.Mission.Segments.Cruise
RCAIDE.Library.Mission.Segments.Ground
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from . import Climb
from . import Hover 
from . import Descent


# RCAIDE/Methods/Performance/__init__.py
# 

"""
Methods for analyzing vehicle performance characteristics including aerodynamics, propulsion, 
and flight mechanics. 
 
This module provides functions for estimating key performance metrics 
such as take-off and landing distances, stall speeds, payload-range capabilities, and flight 
envelope characteristics.
 
See Also
--------
RCAIDE.Library.Mission
RCAIDE.Library.Methods.Aerodynamics
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
from .aircraft_aerodynamic_analysis     import aircraft_aerodynamic_analysis
from .estimate_take_off_field_length    import estimate_take_off_field_length
from .estimate_stall_speed              import estimate_stall_speed
from .compute_payload_range_diagram     import compute_payload_range_diagram
from .estimate_landing_field_length     import estimate_landing_field_length
from .find_take_off_weight_given_tofl   import find_take_off_weight_given_tofl
from .generate_V_n_diagram              import generate_V_n_diagram 
from .rotor_aerodynamic_analysis        import rotor_aerodynamic_analysis  



# RCAIDE/Library/Methods/Powertrain/Propulsors/Internal_Combustion_Engine/unpack_internal_combustion_engine_unknowns.py
# 
# Created:  Jun 2024, M. Clarke   

# ---------------------------------------------------------------------------------------------------------------------- 
#  unpack ice propeller network unknowns 
# ----------------------------------------------------------------------------------------------------------------------  

def unpack_internal_combustion_engine_unknowns(propulsor,segment):  
    """Unpack internal combustion engine unknowns and assigns them to the specfic
    compoment each interation of the mission solver
    """
    engine            = propulsor.engine 
    segment.state.conditions.energy.converters[engine.tag].omega = segment.state.unknowns[propulsor.tag + '_propeller_omega'] 
    return 


# RCAIDE/Methods/Energy/Propulsors/Constant_Speed_ICE_Propulsor/compute_cs_ice_performance.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE imports  
from RCAIDE.Framework.Core import Units  
from RCAIDE.Library.Methods.Powertrain.Converters.Engine import compute_throttle_from_power
from RCAIDE.Library.Methods.Powertrain.Converters.Rotor.compute_rotor_performance import  compute_rotor_performance
 
# pacakge imports  
from copy import deepcopy
import numpy as np  

# ----------------------------------------------------------------------------------------------------------------------
# internal_combustion_engine_constant_speed_propulsor
# ----------------------------------------------------------------------------------------------------------------------  
def compute_constant_speed_internal_combustion_engine_performance(propulsor, state, center_of_gravity=[[0.0, 0.0, 0.0]]):
    """
    Computes the performance of a constant speed internal combustion engine.
    
    Parameters
    ----------
    propulsor : RCAIDE.Library.Components.Propulsors.Constant_Speed_Internal_Combustion_Engine
        Constant speed internal combustion engine component with the following attributes:
            - tag : str
                Identifier for the propulsor
            - engine : Data
                Engine component
                    - tag : str
                        Identifier for the engine
            - propeller : Data
                Propeller component
                    - tag : str
                        Identifier for the propeller
                    - origin : list of lists
                        Origin coordinates [[x, y, z]] [m]
    state : RCAIDE.Framework.Mission.Common.State
        State object containing:
            - conditions : Data
                Flight conditions
                    - energy : Data
                        Energy conditions
                        - propulsors : dict
                            Propulsor energy conditions indexed by tag
                        - converters : dict
                            Converter energy conditions indexed by tag
            - ones_row : function
                Function to create array of ones with specified length
    center_of_gravity : list of lists, optional
        Center of gravity coordinates [[x, y, z]] [m]
        Default: [[0.0, 0.0, 0.0]]
    
    Returns
    -------
    thrust : numpy.ndarray
        Thrust force vector [N]
    moment : numpy.ndarray
        Moment vector [N·m]
    power : numpy.ndarray
        Mechanical power [W]
    power_elec : numpy.ndarray
        Electrical power [W] (zero for non-hybrid engines)
    stored_results_flag : bool
        Flag indicating if results are stored
    stored_propulsor_tag : str
        Tag of the propulsor with stored results
    
    Notes
    -----
    This function computes the performance of a constant speed internal combustion engine
    propulsion system by linking the engine to the propeller. Unlike variable speed engines,
    the constant speed engine maintains a fixed angular velocity regardless of flight condition.
    
    The function performs the following steps:
        - Sets the propeller's angular velocity to match the engine's constant speed
        - Sets the propeller's blade pitch command based on the throttle setting
        - Computes the propeller performance
        - Calculates the moment generated by the propeller thrust
        - Computes the engine power required to match the propeller power demand
        - Calculates the fuel flow rate based on the power demand
        - Stores the results in the conditions data structure
    
    The constant speed is maintained by varying the propeller blade pitch rather than
    the engine RPM, which is common in generator applications and some aircraft engines.
    
    **Major Assumptions**
        * Engine operates at a constant angular velocity
        * Propeller blade pitch is adjusted to control thrust
        * No electrical power generation or consumption (non-hybrid configuration)
        * Moment is calculated relative to the aircraft center of gravity
    
    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Converters.Engine.compute_throttle_from_power
    RCAIDE.Library.Methods.Powertrain.Converters.Rotor.compute_rotor_performance
    """
    conditions              = state.conditions  
    ice_cs_conditions       = conditions.energy.propulsors[propulsor.tag] 
    engine                  = propulsor.engine 
    propeller               = propulsor.propeller 

    # Run the propeller to get the power 
    conditions.energy.converters[propeller.tag].omega                = conditions.energy.converters[engine.tag].omega  
    conditions.energy.converters[propeller.tag].blade_pitch_command  = ice_cs_conditions.throttle - 0.5
    conditions.energy.converters[propeller.tag].throttle             = ice_cs_conditions.throttle
    compute_rotor_performance(propeller,conditions)

    # Compute moment 
    moment_vector           = 0*state.ones_row(3)
    moment_vector[:,0]      = propeller.origin[0][0]  -  center_of_gravity[0][0] 
    moment_vector[:,1]      = propeller.origin[0][1]  -  center_of_gravity[0][1] 
    moment_vector[:,2]      = propeller.origin[0][2]  -  center_of_gravity[0][2]
    moment                  =  np.cross(moment_vector, conditions.energy.converters[propeller.tag].thrust)       
        

    # Run the engine to calculate the throttle setting and the fuel burn
    conditions.energy.converters[engine.tag].power        = conditions.energy.converters[propeller.tag].power 
    compute_throttle_from_power(engine,conditions) 
    
    # Create the outputs
    ice_cs_conditions.fuel_flow_rate         = conditions.energy.converters[engine.tag].fuel_flow_rate  
    stored_results_flag                      = True
    stored_propulsor_tag                     = propulsor.tag  

    # compute total forces and moments from propulsor (future work would be to add moments from motors)
    ice_cs_conditions.thrust      = conditions.energy.converters[propeller.tag].thrust 
    ice_cs_conditions.moment      = moment
    ice_cs_conditions.power       = conditions.energy.converters[propeller.tag].power  

    # currently, no hybridization
    power_elec =  0*state.ones_row(1)
    
    return ice_cs_conditions.thrust ,ice_cs_conditions.moment,ice_cs_conditions.power,power_elec,stored_results_flag,stored_propulsor_tag 
    
def reuse_stored_constant_speed_internal_combustion_engine_data(propulsor,state,network,stored_propulsor_tag,center_of_gravity= [[0.0, 0.0,0.0]]):
    '''Reuses results from one propulsor for identical propulsors
    
    Assumptions: 
    N/A

    Source:
    N/A

    Inputs:  
    conditions           - operating conditions data structure    [-]  
    fuel_line            - fuel_line                              [-] 
    propulsor            - propulsor data structure     [-] 
    total_thrust         - thrust of propulsor group       [N]
    total_power          - power of propulsor group        [W] 
     
    Outputs:      
    total_thrust         - thrust of propulsor group       [N]
    total_power          - power of propulsor group        [W] 
    
    Properties Used: 
    N.A.        
    '''
    # unpack 
    conditions                 = state.conditions
    engine                     = propulsor.engine
    propeller                  = propulsor.propeller 
    engine_0                   = network.propulsors[stored_propulsor_tag].engine
    propeller_0                = network.propulsors[stored_propulsor_tag].propeller 
    
    # deep copy results
    conditions.energy.propulsors[propulsor.tag]     = deepcopy(conditions.energy.propulsors[stored_propulsor_tag.tag])
    conditions.energy.converters[engine.tag]        = deepcopy(conditions.energy.converters[engine_0.tag])
    conditions.energy.converters[propeller.tag]     = deepcopy(conditions.energy.converters[propeller_0.tag])

    # compute moment    
    thrust                  = conditions.energy.converters[propeller.tag].thrust 
    power                   = conditions.energy.converters[propeller.tag].power  
    moment_vector           = 0*state.ones_row(3) 
    moment_vector[:,0]      = propeller.origin[0][0]  -  center_of_gravity[0][0] 
    moment_vector[:,1]      = propeller.origin[0][1]  -  center_of_gravity[0][1] 
    moment_vector[:,2]      = propeller.origin[0][2]  -  center_of_gravity[0][2]
    moment                  =  np.cross(moment_vector, thrust)
    
    # pack results 
    conditions.energy.converters[propeller.tag].moment = moment  
    conditions.energy.propulsors[propulsor.tag].thrust = thrust   
    conditions.energy.propulsors[propulsor.tag].moment = moment  
    conditions.energy.propulsors[propulsor.tag].power  = power
    
    power_elec =  0*state.ones_row(1)  
    return thrust,moment,power, power_elec
 


# RCAIDE/Library/Methods/Powertrain/Converters/DC_generator/design_optimal_generator.py
# 
# Created:  Jul 2024, RCAIDE Team 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
import RCAIDE

# python imports  
from scipy.optimize import minimize 

# ----------------------------------------------------------------------------------------------------------------------
#  design generator 
# ----------------------------------------------------------------------------------------------------------------------     
def design_optimal_generator(generator):
    """
    Sizes a generator to obtain the best combination of speed constant and resistance values
    by sizing the generator for a design RPM value.
    
    Parameters
    ----------
    generator : RCAIDE.Library.Components.Powertrain.Converters.DC_Generator
        Generator component with the following attributes:
            - no_load_current : float
                No-load current [A]
            - nominal_voltage : float
                Nominal voltage [V]
            - angular_velocity : float
                Angular velocity [radians/s]
            - efficiency : float
                Efficiency [unitless]
            - design_torque : float
                Design torque [N·m]
            - gearbox : Data
                Gearbox component
                    - gear_ratio : float
                        Gear ratio [unitless]
            - design_angular_velocity : float
                Design angular velocity [radians/s]
            - design_power : float
                Design power [W]
    
    Returns
    -------
    generator : RCAIDE.Library.Components.Powertrain.Converters.DC_Generator
        Generator with updated attributes:
            - speed_constant : float
                Speed constant [unitless]
            - resistance : float
                Resistance [ohms]
    
    Notes
    -----
    This function uses numerical optimization to find the optimal values of speed constant
    and resistance that satisfy the generator's design requirements. It attempts to solve
    the optimization problem with hard constraints first, and if that fails, it uses slack
    constraints.
    
    The optimization process follows these steps:
        1. Extract generator design parameters (voltage, angular velocity, efficiency, power)
        2. Define optimization bounds for speed constant and resistance
        3. Set up hard constraints for efficiency and power
        4. Attempt optimization with hard constraints
        5. If optimization fails, retry with slack constraints
        6. Update the generator with the optimized parameters
    
    The objective function maximizes the current output for a given voltage and angular velocity.
    
    **Major Assumptions**
        * The generator follows a DC generator model
        * The optimization bounds are appropriate for the generator size
        * If hard constraints cannot be satisfied, slack constraints are acceptable
    
    **Theory**
    The generator model uses the following relationships:
        - Current: :math:`I = (V - \\omega/Kv)/R - I₀`
        - Efficiency: :math:`\\eta = (1 - (I₀\\cdot R)/(V - \\omega/Kv))\\cdot(\\omega/(V\\cdot Kv))`
        - Power: :math:`P = \\omega\\cdot I/Kv`
    
    where:
        - V is the nominal voltage
        - ω is the angular velocity
        - Kv is the speed constant
        - R is the resistance
        - I₀ is the no-load current
        - η is the efficiency
        - P is the power
    
    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Converters.Generator.compute_generator_performance
    """
    
    if type(generator) != RCAIDE.Library.Components.Powertrain.Converters.DC_Generator:
        raise Exception('function only supports low-fidelity (DC) generator')
    
    # design properties of the generator 
    io     = generator.no_load_current
    v      = generator.nominal_voltage  
    G      = generator.gearbox.gear_ratio      
    omega  = generator.design_angular_velocity /G     
    etam   = generator.efficiency 
    P      = generator.design_power
    
    # define optimizer bounds 
    KV_lower_bound  = 0.01
    KV_upper_bound  = 100
    Res_lower_bound = 0.001
    Res_upper_bound = 10
    
    args       = (v , omega,  etam , P , io ) 
    hard_cons  = [{'type':'eq', 'fun': hard_constraint_1,'args': args},{'type':'eq', 'fun': hard_constraint_2,'args': args}] 
    slack_cons = [{'type':'eq', 'fun': slack_constraint_1,'args': args},{'type':'eq', 'fun': slack_constraint_2,'args': args}]  
    bnds       = ((KV_lower_bound, KV_upper_bound), (Res_lower_bound , Res_upper_bound)) 
    
    # try hard constraints to find optimum generator parameters
    sol = minimize(objective, [0.5, 0.1], args=(v , omega,  etam , P , io) , method='SLSQP', bounds=bnds, tol=1e-6, constraints=hard_cons) 
    
    if sol.success == False:
        # use slack constraints if optimizer fails and generator parameters cannot be found 
        print('\n Optimum generator design failed. Using slack constraints')
        sol = minimize(objective, [0.5, 0.1], args=(v , omega,  etam , P , io) , method='SLSQP', bounds=bnds, tol=1e-6, constraints=slack_cons) 
        if sol.success == False:
            assert('\n Slack contraints failed')  
    
    generator.speed_constant   = sol.x[0]
    generator.resistance       = sol.x[1]    
    
    return generator  
  
# objective function
def objective(x, v , omega,  etam , P , io ): 
    return (v - omega/x[0])/x[1]   

# hard efficiency constraint
def hard_constraint_1(x, v , omega,  etam , P , io ): 
    return etam - (1- (io*x[1])/(v - omega/x[0]))*(omega/(v*x[0]))   

# hard torque equality constraint
def hard_constraint_2(x, v , omega,  etam , P , io ): 
    P_guess = omega * ((v - omega/x[0])/x[1] - io)/x[0]
    return P_guess - P  

# slack efficiency constraint 
def slack_constraint_1(x, v , omega,  etam , P , io ): 
    return abs(etam - (1- (io*x[1])/(v - omega/x[0]))*(omega/(v*x[0]))) - 0.2

# slack torque equality constraint 
def slack_constraint_2(x, v , omega,  etam , P , io ): 
    P_guess = omega * ((v - omega/x[0])/x[1] - io)/x[0] 
    return  abs(P_guess - P) - 200 


# RCAIDE/Library/Methods/Powertrain/Converters/Generator/__init__.py

# Created:  Heb 2025, M. Clarke 

"""
generator Methods Package

This module contains methods for generator performance analysis and condition handling.
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .compute_generator_performance             import compute_generator_performance 
from .append_generator_conditions               import append_generator_conditions
from .design_optimal_generator                  import design_optimal_generator


# RCAIDE/Library/Methods/Powertrain/Propulsors/Electric_Rotor/design_electric_rotor.py
# 
# Created:  Mar 2025, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE Imports
import RCAIDE 
from RCAIDE.Library.Methods.Powertrain.Converters.Rotor                     import design_propeller 
from RCAIDE.Library.Methods.Powertrain.Converters.Rotor                     import design_lift_rotor 
from RCAIDE.Library.Methods.Powertrain.Converters.Rotor                     import design_prop_rotor 
from RCAIDE.Library.Methods.Powertrain.Converters.Motor                     import design_optimal_motor 
from RCAIDE.Library.Methods.Mass_Properties.Weight_Buildups.Electric.Common import compute_motor_weight
from RCAIDE.Library.Methods.Powertrain                                      import setup_operating_conditions 

# Python package imports
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
#  Design Electric Rotor 
# ---------------------------------------------------------------------------------------------------------------------- 
def design_electric_rotor(electric_rotor, number_of_stations=20, solver_name='SLSQP', iterations=200,
                         solver_sense_step=1E-6, solver_tolerance=1E-5, print_iterations=False):
    """
    Computes performance properties of an electrically powered rotor.
    
    Parameters
    ----------
    electric_rotor : RCAIDE.Library.Components.Propulsors.Electric_Rotor
        Electric rotor propulsor component with the following attributes:
            - tag : str
                Identifier for the propulsor
            - electronic_speed_controller : Data
                ESC component
                    - bus_voltage : float
                        Bus voltage [V]
            - rotor : Data
                Rotor component (Propeller, Lift_Rotor, or Prop_Rotor)
            - motor : Data
                Electric motor component
                    - design_torque : float
                        Design torque [N·m]
                    - design_angular_velocity : float
                        Design angular velocity [rad/s]
                    - design_current : float
                        Design current [A]
    number_of_stations : int, optional
        Number of radial stations for rotor blade discretization
        Default: 20
    solver_name : str, optional
        Name of the numerical solver to use for rotor design
        Default: 'SLSQP'
    iterations : int, optional
        Maximum number of iterations for the solver
        Default: 200
    solver_sense_step : float, optional
        Step size for finite difference approximations in the solver
        Default: 1E-6
    solver_tolerance : float, optional
        Convergence tolerance for the solver
        Default: 1E-5
    print_iterations : bool, optional
        Flag to print solver iterations
        Default: False
    
    Returns
    -------
    None
        Results are stored in the electric_rotor object:
            - sealevel_static_thrust : float
                Sea level static thrust [N]
            - sealevel_static_power : float
                Sea level static power [W]
    
    Notes
    -----
    This function performs several tasks:
        1. Designs the rotor based on its type (propeller, lift rotor, or prop rotor)
        2. Sets the motor design parameters based on the rotor requirements
        3. Designs the motor for optimal performance
        4. Computes the motor weight
        5. Calculates the sea level static performance (thrust and power)
    
    The function handles different types of rotors:
        * For propellers, it uses the design_propeller function and sets the motor design
          parameters based on cruise conditions
        * For prop rotors, it uses the design_prop_rotor function and sets the motor design
          parameters based on hover conditions
        * For lift rotors, it uses the design_lift_rotor function and sets the motor design
          parameters based on hover conditions
    
    **Major Assumptions**
        * US Standard Atmosphere 1976 is used for atmospheric properties
        * Sea level static conditions are approximated with a very low velocity (1% of speed of sound)
        * Full throttle (throttle = 1.0) is used for sea level static performance
    
    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Converters.Rotor.design_propeller
    RCAIDE.Library.Methods.Powertrain.Converters.Rotor.design_lift_rotor
    RCAIDE.Library.Methods.Powertrain.Converters.Rotor.design_prop_rotor
    RCAIDE.Library.Methods.Powertrain.Converters.Motor.design_optimal_motor
    RCAIDE.Library.Methods.Mass_Properties.Weight_Buildups.Electric.Common.compute_motor_weight
    RCAIDE.Library.Methods.Powertrain.setup_operating_conditions
    """

    if electric_rotor.electronic_speed_controller == None: 
        raise AssertionError("Electric Speed Controller not defined on propulsor")
    
    if electric_rotor.electronic_speed_controller.bus_voltage == None: 
        raise AssertionError("Electric Speed Controller  bus voltage not specified on propulsor") 
    
    if electric_rotor.rotor == None:
        raise AssertionError("Rotor not defined on propulsor")
    rotor = electric_rotor.rotor

    if electric_rotor.motor == None:
        raise AssertionError("Motor not defined on propulsor")
    
    motor = electric_rotor.motor
    
    if type(rotor) == RCAIDE.Library.Components.Powertrain.Converters.Propeller: 
        design_propeller(rotor,number_of_stations = number_of_stations)
        motor.design_torque            = rotor.cruise.design_torque 
        motor.design_angular_velocity  = rotor.cruise.design_angular_velocity 
    elif type(rotor) == RCAIDE.Library.Components.Powertrain.Converters.Prop_Rotor:
        design_prop_rotor(rotor,number_of_stations ,solver_name,iterations,solver_sense_step,solver_tolerance,print_iterations)
        motor.design_torque            = rotor.hover.design_torque 
        motor.design_angular_velocity  = rotor.hover.design_angular_velocity 
    elif type(rotor) == RCAIDE.Library.Components.Powertrain.Converters.Lift_Rotor: 
        design_lift_rotor(rotor,number_of_stations ,solver_name,iterations,solver_sense_step,solver_tolerance,print_iterations)
        motor.design_torque            = rotor.hover.design_torque 
        motor.design_angular_velocity  = rotor.hover.design_angular_velocity
    
    # design motor if design torque is specified 
    if motor.design_torque != None: 
        design_optimal_motor(motor)
    
        # compute weight of motor 
        compute_motor_weight(motor) 
     
    # Static Sea Level Thrust   
    atmosphere            = RCAIDE.Framework.Analyses.Atmospheric.US_Standard_1976() 
    atmo_data_sea_level   = atmosphere.compute_values(0.0,0.0)   
    V                     = atmo_data_sea_level.speed_of_sound[0][0]*0.01 
    operating_state       = setup_operating_conditions(electric_rotor, altitude = 0,velocity_range=np.array([V]))  
    operating_state.conditions.energy.propulsors[electric_rotor.tag].throttle[:,0] = 1.0
    operating_state.conditions.energy.converters[motor.tag].inputs.current[:,0] =  motor.design_current
    sls_T,_,sls_P,_,_,_                          = electric_rotor.compute_performance(operating_state) 
    electric_rotor.sealevel_static_thrust        = sls_T[0][0]
    electric_rotor.sealevel_static_power         = sls_P[0][0]
     
    return 


# RCAIDE/Library/Methods/Weights/Correlation_Buildups/Common/compute_payload_weight.py
# 
# Created: Sep 2024, M. Clarke 

# ---------------------------------------------------------------------------------------------------------------------- 
#  Imports
# ----------------------------------------------------------------------------------------------------------------------
from RCAIDE.Framework.Core import Data, Units 

# ---------------------------------------------------------------------------------------------------------------------- 
# Payload
# ---------------------------------------------------------------------------------------------------------------------- 
def compute_payload_weight(vehicle, W_passenger=195 * Units.lbs, W_baggage=30 * Units.lbs):
    """
    Computes the total payload weight including passengers, baggage, and cargo based on 
    FAA standard weights and aircraft configuration.

    Parameters
    ----------
    vehicle : Vehicle
        The vehicle instance containing:
            - passengers : int
                Number of passengers
            - mass_properties.cargo : float
                Mass of cargo [kg]
    W_passenger : float, optional
        Standard passenger weight [kg], default 195 lbs
    W_baggage : float, optional
        Standard baggage weight per passenger [kg], default 30 lbs

    Returns
    -------
    output : Data
        Container with payload breakdown:
            - total : float
                Total payload weight [kg]
            - passengers : float
                Total passenger weight [kg]
            - baggage : float
                Total baggage weight [kg]
            - cargo : float
                Bulk cargo weight [kg]

    Notes
    -----
    Uses FAA standard weights for passengers and baggage in commercial operations.

    **Major Assumptions**
        * Standard passenger weights
        * Fixed baggage allowance per passenger
        * Uniform passenger distribution
        * No special cargo requirements
        * No seasonal weight variations

    **Theory**
    Total payload weight is computed as:
    .. math::
        W_{payload} = n_{pax}(W_{pax} + W_{bag}) + W_{cargo}

    where:
        * n_pax = number of passengers
        * W_pax = standard passenger weight
        * W_bag = standard baggage allowance
        * W_cargo = bulk cargo weight

    See Also
    --------
    RCAIDE.Library.Methods.Mass_Properties.Weight_Buildups.Conventional
    """

    # process
    num_pax    = vehicle.passengers
    W_pax      = W_passenger * num_pax
    W_bag      = W_baggage * num_pax
    if vehicle.mass_properties.payload == 0:
        vehicle.mass_properties.payload  = W_pax + W_bag + vehicle.mass_properties.cargo
    else:
        vehicle.mass_properties.cargo = vehicle.mass_properties.payload - W_pax - W_bag

    # packup outputs
    output              = Data()
    output.total        = vehicle.mass_properties.payload
    output.passengers   = W_pax
    output.baggage      = W_bag
    output.cargo        = vehicle.mass_properties.cargo

    return output


# RCAIDE/Library/Methods/Powertrain/Converters/Turboshaft/append_turboshaft_conditions.py
# 
# Created:  Jun 2024, M. Clarke  
# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE imports  
import RCAIDE
from RCAIDE.Framework.Mission.Common     import   Conditions

# ---------------------------------------------------------------------------------------------------------------------- 
#  append_turboshaft_conditions
# ----------------------------------------------------------------------------------------------------------------------    
def append_turboshaft_conditions(turboshaft, segment, energy_conditions, noise_conditions):
    """
    Appends data structures for storing turboshaft operating conditions during mission analysis.
    
    Parameters
    ----------
    turboshaft : RCAIDE.Components.Energy.Converters.Turboshaft
        The turboshaft component for which conditions are being appended
    segment : RCAIDE.Analyses.Mission.Segments
        The mission segment being evaluated
    energy_conditions : RCAIDE.Framework.Mission.Common.Conditions
        Container for energy-related conditions during the mission segment
    noise_conditions : RCAIDE.Framework.Mission.Common.Conditions
        Container for noise-related conditions during the mission segment
        
    Returns
    -------
    None
        This function modifies the energy_conditions and noise_conditions objects in-place
    
    Notes
    -----
    This function initializes the condition structure for a turboshaft engine with zero values
    for throttle, thrust vector angle, power, and fuel flow rate. It then recursively calls
    the append_operating_conditions method for all subcomponents of the turboshaft.
    
    The function creates a hierarchical data structure that mirrors the component hierarchy
    of the turboshaft engine, allowing for detailed tracking of operating conditions at each
    level during mission analysis.
    
    **Major Assumptions**
        * The segment state contains a ones_row method for creating arrays
        * All subcomponents have properly implemented append_operating_conditions methods
    
    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Converters.Turboshaft.compute_turboshaft_performance
    RCAIDE.Library.Components.Powertrain.Converters.Turboshaft
    """
    ones_row    = segment.state.ones_row
    
    energy_conditions.converters[turboshaft.tag]                               = Conditions() 
    energy_conditions.converters[turboshaft.tag].throttle                      = 0. * ones_row(1)     
    energy_conditions.converters[turboshaft.tag].commanded_thrust_vector_angle = 0. * ones_row(1)   
    energy_conditions.converters[turboshaft.tag].power                         = 0. * ones_row(1)
    energy_conditions.converters[turboshaft.tag].fuel_flow_rate                = 0. * ones_row(1)
    energy_conditions.converters[turboshaft.tag].inputs                        = Conditions()
    energy_conditions.converters[turboshaft.tag].outputs                       = Conditions()
 
    for tag, item in  turboshaft.items(): 
        if issubclass(type(item), RCAIDE.Library.Components.Component):
            item.append_operating_conditions(segment,energy_conditions,noise_conditions) 
            for sub_tag, sub_item in  item.items(): 
                if issubclass(type(sub_item), RCAIDE.Library.Components.Component):
                    sub_item.append_operating_conditions(segment,energy_conditions,noise_conditions) 
    return 


# RCAIDE/Library/Missions/Common/Unpack_Unknowns/control_surfaces.py
# 
# 
# Created:  Jul 2023, M. Clarke
import RCAIDE

# ----------------------------------------------------------------------------------------------------------------------
#  Unpack Unknowns
# ----------------------------------------------------------------------------------------------------------------------
def control_surfaces(segment):
    """
    Updates control surface deflections from solver unknowns

    Parameters
    ----------
    segment : Segment
        The mission segment being analyzed
            - assigned_control_variables : Data
                Control variable configurations
                - {control_type}.active : bool
                    Whether control is active
                - {control_type}.assigned_surfaces : list
                    Surface names for each control group
            - state.unknowns : Data
                Solver unknown values
            - state.conditions.control_surfaces : Data
                Results data structure
            - analyses : list
                Analysis modules containing vehicle definition
    
    Returns
    -------
    None
        Updates segment state and vehicle model directly

    Notes
    -----
    This function applies control surface deflection values from the solver's
    unknowns to both the vehicle model and results data structure. It handles
    all types of control surfaces including elevators, slats, rudders, flaps,
    and ailerons.

    The function processes:
        1. Elevator deflections
        2. Slat deflections
        3. Rudder deflections
        4. Flap deflections
        5. Aileron deflections
        6. Spoiler deflections

    **Control Surface Types**
    
    Supported controls:
        - Elevator
        - Slat
        - Rudder
        - Flap
        - Aileron
        - Spoiler

    **Major Assumptions**
        * Valid control surface definitions
        * Proper surface assignments
        * Compatible deflection values
        * Well-defined vehicle geometry

    See Also
    --------
    RCAIDE.Library.Components.Wings.Control_Surfaces
    RCAIDE.Framework.Mission.Segments
    """
    
    
    assigned_control_variables   = segment.assigned_control_variables
    control_surfaces             = segment.state.conditions.control_surfaces
    
    for analysis in segment.analyses:
        if analysis !=  None: 
            if 'vehicle' in analysis: 
                wings =  analysis.vehicle.wings 
                # loop through wings on aircraft
                for wing in wings:
                    # Elevator Control
                    if assigned_control_variables.elevator_deflection.active:
                        for control_surface in wing.control_surfaces:
                            if type(control_surface) == RCAIDE.Library.Components.Wings.Control_Surfaces.Elevator: 
                                control_surfaces.elevator.deflection  = segment.state.unknowns["elevator"]
                            
                    # Slat Control
                    if assigned_control_variables.slat_deflection.active:
                        for control_surface in wing.control_surfaces:
                            if type(control_surface) == RCAIDE.Library.Components.Wings.Control_Surfaces.Slat: 
                                control_surfaces.slat.deflection  = segment.state.unknowns["slat"]

                    # Rudder Control
                    if assigned_control_variables.rudder_deflection.active:
                        for control_surface in wing.control_surfaces:
                            if type(control_surface) == RCAIDE.Library.Components.Wings.Control_Surfaces.Rudder: 
                                control_surfaces.rudder.deflection  = segment.state.unknowns["rudder"]
            
                    # flap Control
                    if assigned_control_variables.flap_deflection.active:
                        for control_surface in wing.control_surfaces:
                            if type(control_surface) == RCAIDE.Library.Components.Wings.Control_Surfaces.Flap: 
                                control_surfaces.flap.deflection  = segment.state.unknowns["flap"]
            
                    # Aileron Control
                    if assigned_control_variables.aileron_deflection.active:
                        for control_surface in wing.control_surfaces:
                            if type(control_surface) == RCAIDE.Library.Components.Wings.Control_Surfaces.Aileron: 
                                control_surfaces.aileron.deflection  = segment.state.unknowns["aileron"]
    return


# RCAIDE/Library/Attributes/Gases/CO2.py
# 
#
# Created:  Mar 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------  
#  Imports
# ----------------------------------------------------------------------------------------------------------------------  
 
from .Gas import Gas  

# ----------------------------------------------------------------------------------------------------------------------  
# Steam Class
# ----------------------------------------------------------------------------------------------------------------------   
class Steam(Gas):
    """
    A class representing water vapor (steam) and its thermodynamic properties. Provides methods for computing 
    various gas properties including density, speed of sound, and specific heat.

    Attributes
    ----------
    tag : str
        Identifier for the gas type ('steam')
    
    molecular_mass : float
        Molecular mass of H2O in kg/kmol (18.0)
    
    gas_specific_constant : float
        Specific gas constant in m²/s²-K (461.889)
    
    composition : Data
        Chemical composition of steam
        
        - H2O : float
            Mass fraction of water vapor (1.0)

    Notes
    -----
    This class implements steam properties assuming ideal gas behavior for most calculations.
    All properties are for pure water vapor.
    """

    def __defaults__(self):
        """This sets the default values.
    
        Assumptions:
            None
        
        Source:
            None
        """ 
        self.tag                   ='steam'
        self.molecular_mass        = 18.           # kg/kmol
        self.gas_specific_constant = 461.889       # m^2/s^2-K, specific gas constant
        self.composition.H2O       = 1.0

    def compute_density(self,T=300,p=101325):
        """
        Computes steam density using the ideal gas law.

        Parameters
        ----------
        T : float
            Temperature in Kelvin
        p : float
            Pressure in Pascal

        Returns
        -------
        rho : float
            Density in kg/m³

        Notes
        -----
        **Major Assumptions**
        * Steam behaves as an ideal gas   
        """        
        return p/(self.gas_specific_constant*T)

    def compute_speed_of_sound(self,T=300,p=101325,variable_gamma=False):
        """
        Computes speed of sound in steam.

        Parameters
        ----------
        T : float
            Temperature in Kelvin
        p : float
            Pressure in Pascal
        variable_gamma : bool
            If True, uses temperature-dependent specific heat ratio

        Returns
        -------
        a : float
            Speed of sound in m/s

        Notes
        -----
        **Major Assumptions**
            * If variable_gamma is False, assumes γ = 1.33
            * Steam behaves as an ideal gas
        """        
        if variable_gamma:
            g = self.compute_gamma(T,p)
        else:
            g = 1.33

        return (g*self.gas_specific_constant*T)**0.5 

    def compute_cp(self,T=300,p=101325):
        """
        Computes specific heat capacity at constant pressure using a 3rd-order polynomial fit.

        Parameters
        ----------
        T : float
            Temperature in Kelvin
        p : float
            Pressure in Pascal

        Returns
        -------
        cp : float
            Specific heat capacity in J/(kg·K)

        Notes
        -----
        **Major Assumptions**
            * Valid for temperature range: 300 K < T < 1500 K

        **Theory**
        .. math::
            c_p(T) = c_1T^3 + c_2T^2 + c_3T + c_4

        References
        ----------
        [1] Carvalho, M. G., Lockwood, F. C., Fiveland, W. A., & Papadopoulos, C. (2022). Combustion Technologies for a clean environment: Selected papers from the proceedings of the first international conference, Vilamoura, Portugal, September 3-6, 1991. CRC Press.       
        """   
        c = [5E-9, -.0001,  .9202, 1524.7]
        cp = c[0]*T**3 + c[1]*T**2 + c[2]*T + c[3]

        return cp

    def compute_gamma(self,T=300,p=101325):
        """
        Returns the specific heat ratio for steam.

        Parameters
        ----------
        T : float
            Temperature in Kelvin
        p : float
            Pressure in Pascal

        Returns
        -------
        g : float
            Specific heat ratio (gamma) [unitless]

        Notes
        -----
        **Major Assumptions**
            * Valid for temperature range: 233 K < T < 1273 K
            * Uses constant value of 1.33
        """           
        g = 1.33   
        return g

    def compute_absolute_viscosity(self,T=300,p=101325):
        """
        Returns the absolute (dynamic) viscosity of steam.

        Parameters
        ----------
        T : float
            Temperature in Kelvin
        p : float
            Pressure in Pascal

        Returns
        -------
        mu : float
            Absolute viscosity in kg/(m·s)

        Notes
        -----
        **Major Assumptions**
            * Uses constant value of 1e-6 kg/(m·s)     
        """ 
        mu =1E-6 
        return mu 


# RCAIDE/Library/Methods/Weights/Correlation_Buildups/FLOPS/compute_fuselage_weight.py
# 
# 
# Created:  Sep 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE
import RCAIDE 
from RCAIDE.Framework.Core    import Units

# python imports 
import  numpy as  np
 
# ----------------------------------------------------------------------------------------------------------------------
# Fuselage Weight 
# ----------------------------------------------------------------------------------------------------------------------
def compute_fuselage_weight(vehicle):
    """
    Calculate the weight of a general aviation aircraft fuselage using NASA FLOPS methodology.

    Parameters
    ----------
    vehicle : RCAIDE.Vehicle()
        Vehicle data structure containing:
            - fuselages.lengths.total : float
                Total length of the fuselage [m]
            - fuselages.width : float
                Maximum width of the fuselage [m]
            - fuselages.heights.maximum : float
                Maximum height of the fuselage [m]
            - flight_envelope.ultimate_load : float
                Ultimate load factor
            - mass_properties.max_takeoff : float
                Maximum takeoff weight [kg]
            - design_dynamic_pressure : float
                Design dynamic pressure [Pa]

    Returns
    -------
    fuselage_weight : float
        Weight of the fuselage structure [kg]

    Notes
    -----
    The function implements the FLOPS (Flight Optimization System) weight estimation
    method for general aviation aircraft fuselages. The calculation accounts for geometric 
    parameters, design loads, and cruise conditions.
    
    **Major Assumptions**
        * Single fuselage configuration (NFUSE = 1)
        * Structural weight scales with wetted area and design loads
    
    **Theory**

    The FLOPS fuselage weight estimation follows:

    .. math::
        W_{fus} = 0.052 * S_{wf}^{1.086} * (N_{ult} * W_{TO})^{0.177} * q^{0.241}

    Where:
        - W_{fus} is fuselage weight [lb]
        - S_{wf} is fuselage wetted area [ft^2], calculated as π(L/D - 1.7)D^2
        - N_{ult} is ultimate load factor
        - W_{TO} is maximum takeoff weight [lb]
        - q is design dynamic pressure [psf]
        - L is fuselage length [ft]
        - D is average fuselage diameter [ft]

    References
    ----------
    [1] NASA. The Flight Optimization System Weights Estimation Method. 
        NASA Technical Report.
    """
    
    L =  0
    for fuselage in vehicle.fuselages:
        if L < fuselage.lengths.total: 
            total_length = fuselage.lengths.total
            width        = fuselage.width
            max_height   = fuselage.heights.maximum
    
    XL  = total_length / Units.ft  # Fuselage length, ft
    DAV = (width + max_height) / 2. * 1 / Units.ft
    
    DG       = vehicle.mass_properties.max_takeoff / Units.lbs  # Design gross weight in lb
    QCRUS    = vehicle.flight_envelope.design_dynamic_pressure / Units.psf
    ULF      = vehicle.flight_envelope.ultimate_load  
    SWFUS    = 3.14159*(XL/DAV -1.7)*(DAV**2)

    WFUSE = 0.052 * SWFUS**1.086 * (ULF*DG)**0.177 * QCRUS**0.241 
    
    return WFUSE * Units.lbs


# RCAIDE/Library/Methods/Weights/Correlation_Buildups/Propulsion/compute_jet_engine_weight.py
# 
# 
# Created:  Sep 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE 
import RCAIDE
from RCAIDE.Framework.Core import  Units, Data
import numpy as np
 

def compute_propulsion_system_weight(network):
    W_energy_network_total = 0
    number_of_jet_engines = 0
    number_of_piston_engines = 0
    for propulsor in network.propulsors: # Check this 
            if isinstance(propulsor, RCAIDE.Library.Components.Powertrain.Propulsors.Turbofan):
                number_of_jet_engines += 1
                W_engine_jet            = compute_jet_engine_weight(propulsor)
                W_propulsion            = integrated_propulsion_jet(W_engine_jet) 
                propulsor.mass_properties.mass = W_propulsion
                W_energy_network_total  += W_propulsion                
            elif isinstance(propulsor, RCAIDE.Library.Components.Powertrain.Propulsors.Internal_Combustion_Engine):    
                number_of_piston_engines += 1
                W_engine_piston          = compute_piston_engine_weight(propulsor)
                W_propulsion             = integrated_propulsion_piston(W_engine_piston) 
                propulsor.mass_properties.mass = W_propulsion
                W_energy_network_total  += W_propulsion
            elif type(propulsor) ==  RCAIDE.Library.Components.Powertrain.Propulsors.Turboprop:      
                number_of_piston_engines += 1
                W_turboprop          = compute_turboprop_engine_weight(propulsor)
                W_propulsion             = integrated_propulsion_piston(W_turboprop) 
                propulsor.mass_properties.mass = W_propulsion
                W_energy_network_total  += W_propulsion     
    
    output = Data()
    output.W_prop = W_propulsion
    output.number_of_engines = number_of_jet_engines + number_of_piston_engines
    return output   
                

# ----------------------------------------------------------------------------------------------------------------------
#  Jet Engine Weight 
# ---------------------------------------------------------------------------------------------------------------------- 
def compute_jet_engine_weight(propulsor):
    """ Calculate the weight of the dry engine  
    
    Assumptions:
            calculated engine weight from a correlation of engines
    
    Source: 
            N/A
            
    Inputs:
            thrust_sls - sea level static thrust of a single engine [Newtons]
    
    Outputs:
            weight - weight of the dry engine                       [kilograms]
        
    Properties Used:
            N/A
    """     
    # setup
    thrust_sls    =  propulsor.sealevel_static_thrust 
    thrust_sls_en = thrust_sls / Units.force_pound # Convert N to lbs force  
    BPR = propulsor.bypass_ratio

    WENG   = 0.084 *  (propulsor.sealevel_static_thrust/Units.lbf)**1.1 * np.exp(-0.045*BPR) * Units.lbs # Raymer 3rd Edition eq. 10.4 
    
    return WENG
 
# ----------------------------------------------------------------------------------------------------------------------
#  Motor Weight 
# ----------------------------------------------------------------------------------------------------------------------
def compute_motor_weight(motor):
    """ Calculate the weight of motor using NASA correlations 
             
    
    Inputs:
            torque- maximum torque the motor can deliver safely      [N-m]
            kwt2
            xwt
            
    Outputs:
            mass- mass of the motor                                [kilograms]
        
    Properties Used:
            N/A
    """
    
    torque =  motor.design_torque
    mass  = -2E-7 * (torque ** 2) +  0.0117 * torque +  34.124
     
    return mass 

# ----------------------------------------------------------------------------------------------------------------------
#  Piston Engine Weight 
# ----------------------------------------------------------------------------------------------------------------------
def compute_piston_engine_weight(propulsor):
    """ Calculate the weight of an piston engine  
        weight correlation; weight=kwt2*(max_power**xwt)
        Inputs:
                max_power- maximum power the motor can deliver safely [Watts]
                kwt2
                xwt
                
        Outputs:
                weight- weight of the motor [kilograms]
        
        Source: Raymer, Aircraft Design, a Conceptual Approach         
    """    
    bhp    = propulsor.engine.sea_level_power/Units.horsepower
    weight = 5.22*((bhp)**0.780)  # weight in lbs.
    mass   = weight*Units.lbs
    return mass

# ----------------------------------------------------------------------------------------------------------------------
#  Turboprop Engine Weight 
# ----------------------------------------------------------------------------------------------------------------------
def compute_turboprop_engine_weight(propulsor):
    """ Calculate the weight of an turboprop engine  
        weight correlation; weight=kwt2*(max_power**xwt)
        Inputs:
                max_power- maximum power the motor can deliver safely [Watts]
                kwt2
                xwt
                
        Outputs:
                weight- weight of the motor [kilograms]
        
        Source: Raymer, Aircraft Design, a Conceptual Approach         
    """    
    bhp    = propulsor.design_power / Units.horsepower
    weight = 1.67*((bhp)**0.803)  # weight in lbs.
    mass   = weight*Units.lbs
    return mass

# ----------------------------------------------------------------------------------------------------------------------
#  Integrated Propulsion Weight 
# ----------------------------------------------------------------------------------------------------------------------
def integrated_propulsion_jet(jet_engine_weight, engine_W_factor = 1.6):
    """ Calculate the weight of the entire propulsion system 
    
    Assumptions:
            The propulsion system is a fixed 60% greater than the dry engine alone. 
            The propulsion system includes the engines, engine exhaust, reverser, starting,
            controls, lubricating, and fuel systems. The nacelle and pylon weight are also
            part of this calculation.           
            
    Source: 
            N/A
            
    Inputs:
            compute_jet_engine_weight - dry weight of the engine                                             [kilograms]
            num_eng - total number of engines on the aircraft                                 [dimensionless]
            engine_W_factor - weight increase factor for entire integrated propulsion system [dimensionless]
    
    Outputs:
            weight - weight of the full propulsion system                                     [kilograms]
        
    Properties Used:
            N/A
    """   
    
    weight = jet_engine_weight * engine_W_factor
    
    return weight
    
# ----------------------------------------------------------------------------------------------------------------------
#  Integrated Propulsion Weight 
# ----------------------------------------------------------------------------------------------------------------------
def integrated_propulsion_piston(engine_weight):
    """ 
        Calculate the weight of the entire propulsion system        

        Source:
                Source: Raymer, Aircraft Design, a Conceptual Approach        
                
        Inputs:
                piston_engine_weight - dry weight of a single engine                                     [kilograms]
                num_eng - total number of engines on the aircraft                                 [dimensionless]
                engine_W_factor - weight increase factor for entire integrated propulsion system [dimensionless]
        
        Outputs:
                weight - weight of the full propulsion system [kilograms]
    """     
    engine_dry = engine_weight/Units.lbs
    weight     = 2.575 * (engine_dry**0.922)
    mass       = weight*Units.lbs #convert to kg

    return mass


# Bidirectional_Carbon_Fiber.py
#
# Created: Jul, 2017, J. Smart
# Modified: Apr, 2018, J. Smart

#-------------------------------------------------------------------------------
# Imports
#-------------------------------------------------------------------------------

from .Solid import Solid
from RCAIDE.Framework.Core import Units

#-------------------------------------------------------------------------------
# Bi-Directional Carbon Fiber Solid Class
#-------------------------------------------------------------------------------

class Bidirectional_Carbon_Fiber(Solid):
    """ 
    A class representing bidirectional carbon fiber composite material properties.

    Attributes
    ----------
    ultimate_tensile_strength : float
        Maximum tensile stress before failure in Pa (600e6)
    ultimate_shear_strength : float
        Maximum shear stress before failure in Pa (90e6)
    ultimate_bearing_strength : float
        Maximum bearing stress before failure in Pa (600e6)
    yield_tensile_strength : float
        Stress at which material begins to deform plastically in Pa (600e6)
    yield_shear_strength : float
        Shear stress at which material begins to deform plastically in Pa (90e6)
    yield_bearing_strength : float
        Bearing stress at which material begins to deform plastically in Pa (600e6)
    minimum_gage_thickness : float
        Minimum manufacturable thickness in m (420e-6)
    density : float
        Material density in kg/m³ (1600)

    Notes
    -----
    This class implements material properties for bidirectional woven carbon fiber composites.
    Properties are based on median values from manufacturer reported data.
    The equal ultimate and yield strengths reflect the generally linear elastic behavior 
    of carbon fiber composites up to failure.

    **Definitions**
    
    'Bidirectional Carbon Fiber'
        A composite material with carbon fibers woven in two perpendicular directions,
        providing similar properties in both directions
    
    'Ultimate Strength'
        The maximum stress that a material can withstand before failure
    
    References
    ----------
    [1] MatWeb. (n.d.). Overview of materials for Epoxy/Carbon Fiber Composite. Overview of materials for epoxy/carbon fiber composite. https://www.matweb.com/search/datasheet.aspx?matguid=39e40851fc164b6c9bda29d798bf3726 
    """

    def __defaults__(self):
        """Sets material properties at instantiation.

        Assumptions:
        None

        Source:
        N/A

        Inputs:
        N/A

        Outputs:
        N/A

        Properties Used:
        None
        """    

        self.ultimate_tensile_strength  = 600e6     * Units.Pa
        self.ultimate_shear_strength    = 90e6      * Units.Pa
        self.ultimate_bearing_strength  = 600e6     * Units.Pa
        self.yield_tensile_strength     = 600e6     * Units.Pa
        self.yield_shear_strength       = 90e6      * Units.Pa
        self.yield_bearing_strength     = 600e6     * Units.Pa
        self.minimum_gage_thickness     = 420e-6    * Units.m
        self.density                    = 1600.     * Units['kg/(m**3)']



# RCAIDE/Methods/Powertrain/Sources/Batteries/Common/size_module_from_mass.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  METHOD
# ---------------------------------------------------------------------------------------------------------------------- -Common
def size_module_from_mass(battery_module):
    """
    Calculates the maximum energy and power of a battery module based on its mass.
    
    Parameters
    ----------
    battery_module : BatteryModule
        The battery module with the following attributes:
            - mass_properties.mass : float
                Total mass of the battery module [kg]
            - BMS_additional_weight_factor : float
                Factor accounting for battery management system weight
            - cell.mass : float or None
                Mass of a single cell [kg]
            - cell.specific_energy : float
                Specific energy of the cell [J/kg]
            - cell.specific_power : float
                Specific power of the cell [W/kg]
            - maximum_voltage : float
                Maximum voltage of the battery module [V]
            - cell.maximum_voltage : float
                Maximum voltage of a single cell [V]
    
    Returns
    -------
    None
        This function modifies the battery_module object in-place, setting the following attributes:
            - maximum_energy : float
                Maximum energy capacity of the battery module [J]
            - maximum_power : float
                Maximum power output of the battery module [W]
            - initial_maximum_energy : float
                Initial maximum energy capacity (same as maximum_energy) [J]
            - electrical_configuration.series : int
                Number of cells in series
            - electrical_configuration.parallel : int
                Number of cells in parallel
    
    Notes
    -----
    This function calculates the energy and power capabilities of a battery module based on
    its mass and the specific energy/power characteristics of its cells. It also determines
    the electrical configuration (series/parallel arrangement) of the cells.
    
    The function first calculates the effective mass of the cells by removing the mass
    contribution of the battery management system (BMS). It then uses this mass along with
    the specific energy and power values to determine the module's capabilities.
    
    If the cell mass is provided, the function calculates the number of cells that can fit
    within the module mass and arranges them in a series-parallel configuration based on
    the voltage requirements. If cell mass is not provided, a default configuration of
    1 series × 1 parallel is used.
    
    **Major Assumptions**
        * Constant specific energy and power values
        * BMS weight is a fixed fraction of the total module weight
        * All cells have identical characteristics
    
    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Sources.Batteries.Common.compute_module_properties
    """
    mass = battery_module.mass_properties.mass/battery_module.BMS_additional_weight_factor
    
    if battery_module.cell.mass == None: 
        n_series   = 1
        n_parallel = 1 
    else:
        n_cells    = int(mass/battery_module.cell.mass)
        n_series   = int(battery_module.maximum_voltage/battery_module.cell.maximum_voltage)
        n_parallel = int(n_cells/n_series)
        
    battery_module.maximum_energy                    = mass*battery_module.cell.specific_energy  
    battery_module.maximum_power                     = mass*battery_module.cell.specific_power
    battery_module.initial_maximum_energy            = battery_module.maximum_energy    
    battery_module.electrical_configuration.series   = n_series
    battery_module.electrical_configuration.parallel = n_parallel     



# RCAIDE/Methods/Powertrain/Sources/__init__.py
# 

"""
This module provides functionality for modeling energy sources in powertrains. It includes methods for 
batteries, cryogenic tanks, and fuel tanks that store and provide energy for propulsion systems.

See Also
--------
RCAIDE.Library.Methods.Powertrain.Converters
RCAIDE.Library.Methods.Powertrain.Distributors
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from . import Batteries
from . import Cryogenic_Tanks
from . import Fuel_Tanks


# RCAIDE/Library/Plots/Noise/plot_noise_level.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------   
from RCAIDE.Framework.Core import Units
from RCAIDE.Library.Plots.Common import set_axes, plot_style 

import matplotlib.pyplot as plt
import matplotlib.cm as cm
import numpy as np 

# ----------------------------------------------------------------------------------------------------------------------
#  PLOTS
# ---------------------------------------------------------------------------------------------------------------------- 
def plot_noise_level(noise_data,
                     noise_level = False,
                     save_figure = False,
                     save_filename = "Sideline_Noise_Levels",
                     file_type = ".png",
                     width = 11, height = 7):
    """
    Creates a visualization of A-weighted Sound Pressure Levels at various sideline distances.

    Parameters
    ----------
    noise_data : NoiseData
        RCAIDE noise data structure containing:
            - microphone_y_resolution : int
                Number of sideline measurement positions
            - microphone_locations : ndarray
                3D array of microphone positions where:
                    - [:,:,0] : x-positions (longitudinal)
                    - [:,:,1] : y-positions (sideline)
                    - [:,:,2] : z-positions (vertical)
            
    noise_level : ndarray, optional
        2D array of noise levels at measurement points (default: False)
        
    save_figure : bool, optional
        Flag for saving the figure (default: False)
        
    save_filename : str, optional
        Name of file for saved figure (default: "Sideline_Noise_Levels")
        
    file_type : str, optional
        File extension for saved figure (default: ".png")
        
    width : float, optional
        Figure width in inches (default: 11)
        
    height : float, optional
        Figure height in inches (default: 7)

    Returns
    -------
    fig : matplotlib.figure.Figure
        Handle to the generated figure showing noise levels vs range

    Notes
    -----
    Creates visualization showing:
        * Noise level variation with distance
        * Multiple sideline measurement positions
        * Color-coded sideline distances
        * Customizable appearance
    
    **Major Assumptions**
        * Noise levels are A-weighted SPL in dBA
        * Microphones are in a regular grid
        * Range is in nautical miles
        * Sideline distances are in meters
    
    **Definitions**
    
    'Sound Pressure Level'
        A-weighted acoustic pressure level in dBA
    'Sideline Distance'
        Perpendicular distance from flight path
    'Range'
        Distance along flight path
    
    See Also
    --------
    RCAIDE.Library.Plots.Noise.plot_2D_noise_contour : 2D contour visualization
    RCAIDE.Library.Plots.Noise.plot_3D_noise_contour : 3D noise field visualization
    """      
    N_gm_y       = noise_data.microphone_y_resolution 
    gm           = noise_data.microphone_locations    
    gm_x         = gm[:,:,0]
    gm_y         = gm[:,:,1]    
    

    # get plotting style 
    ps      = plot_style()  

    parameters = {'axes.labelsize': ps.axis_font_size,
                  'xtick.labelsize': ps.axis_font_size,
                  'ytick.labelsize': ps.axis_font_size,
                  'axes.titlesize': ps.title_font_size}
    plt.rcParams.update(parameters) 
      
    fig         = plt.figure(save_filename)
    fig.set_size_inches(width,height)
    axes        = fig.add_subplot(1,1,1) 
    
    # get line colors for plots 
    line_colors   = cm.inferno(np.linspace(0,0.9,N_gm_y))  
      
    for k in range(N_gm_y):    
        axes.plot(gm_x[:,0]/Units.nmi, noise_level[:,k], marker = 'o', color = line_colors[k], label= r'mic at y = ' + str(round(gm_y[0,k],1)) + r' m' ) 
    axes.set_ylabel('SPL [dBA]')
    axes.set_xlabel('Range [nmi]')  
    set_axes(axes)
    axes.legend(loc='upper right')         
    if save_figure:
        plt.savefig(save_filename + ".png")    
        
    return fig


# RCAIDE/Library/Methods/Powertrain/Converters/Combustor/compute_combustor_performance.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke


# ---------------------------------------------------------------------------------------------------------------------- 
#  Imports 
# ----------------------------------------------------------------------------------------------------------------------   
import  numpy as  np 

# ---------------------------------------------------------------------------------------------------------------------- 
# compute_combustor_performance
# ----------------------------------------------------------------------------------------------------------------------    
def compute_combustor_performance(combustor, conditions):
    """
    Computes the thermodynamic performance of a combustor in a gas turbine engine.
    
    Parameters
    ----------
    combustor : RCAIDE.Library.Components.Converters.Combustor
        Combustor component with the following attributes:
            - tag : str
                Identifier for the combustor
            - working_fluid : Data
                Working fluid properties object
            - turbine_inlet_temperature : float
                Target turbine inlet temperature [K]
            - pressure_ratio : float
                Pressure ratio across the combustor (typically < 1.0 due to losses)
            - efficiency : float
                Combustion efficiency
            - area_ratio : float
                Exit to inlet area ratio
            - fuel_data : Data
                Fuel properties
                    - specific_energy : float
                        Fuel specific energy [J/kg]
    conditions : RCAIDE.Framework.Mission.Common.Conditions
        Flight conditions with:
            - energy : Data
                Energy conditions
                    - converters : dict
                        Converter energy conditions indexed by tag
    
    Returns
    -------
    None
        Results are stored in conditions.energy.converters[combustor.tag].outputs:
            - stagnation_temperature : numpy.ndarray
                Stagnation temperature at combustor exit [K]
            - stagnation_pressure : numpy.ndarray
                Stagnation pressure at combustor exit [Pa]
            - stagnation_enthalpy : numpy.ndarray
                Stagnation enthalpy at combustor exit [J/kg]
            - fuel_to_air_ratio : numpy.ndarray
                Fuel-to-air ratio
            - static_temperature : numpy.ndarray
                Static temperature at combustor exit [K]
            - static_pressure : numpy.ndarray
                Static pressure at combustor exit [Pa]
            - mach_number : numpy.ndarray
                Mach number at combustor exit
    
    Notes
    -----
    This function computes the thermodynamic properties at the combustor exit based on
    the inlet conditions, combustor characteristics, and fuel properties. It calculates
    the fuel-to-air ratio required to achieve the specified turbine inlet temperature,
    accounting for combustion efficiency and pressure losses.
    
    The computation follows these steps:
        1. Extract inlet conditions (temperature, pressure, Mach number)
        2. Compute working fluid properties (gamma, Cp)
        3. Calculate stagnation pressure at exit using pressure ratio
        4. Set exit stagnation temperature to the specified turbine inlet temperature
        5. Compute stagnation enthalpies at inlet and exit
        6. Calculate fuel-to-air ratio required to achieve the temperature rise
        7. Compute exit static conditions (temperature, pressure) based on exit Mach number
        8. Store all results in the conditions data structure
    
    **Major Assumptions**
        * Constant efficiency and pressure ratio
        * Turbine inlet temperature is controlled to a specified value
        * Mach number is preserved from inlet to exit
    
    References
    ----------
    [1] Cantwell, B., "AA283 Course Notes", Stanford University https://web.stanford.edu/~cantwell/AA283_Course_Material/
    
    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Converters.Turbine.compute_turbine_performance
    """
    combustor_conditions    = conditions.energy.converters[combustor.tag]
    T0                      = combustor_conditions.inputs.static_temperature
    P0                      = combustor_conditions.inputs.static_pressure  
    M0                      = combustor_conditions.inputs.mach_number 
                                 
    # Unpack ram inputs         
    working_fluid           = combustor.working_fluid
 
    # Compute the working fluid properties 
    gamma  = working_fluid.compute_gamma(T0,P0) 
    Cp     = working_fluid.compute_cp(T0,P0) 
    
    # unpacking the values form inputs
    Tt_in    = combustor_conditions.inputs.stagnation_temperature
    Pt_in    = combustor_conditions.inputs.stagnation_pressure
    nondim_r = combustor_conditions.inputs.nondim_mass_ratio 
    Tt4      = combustor.turbine_inlet_temperature *  np.ones_like(Tt_in)
    pib      = combustor.pressure_ratio
    eta_b    = combustor.efficiency
    htf      = combustor.fuel_data.specific_energy 
    
    # compute stanation pressure 
    Pt_out  = pib * Pt_in 
    
    #Computing stagnation enthalpies from stagnation temperatures
    ht4     = nondim_r * Cp* Tt4 
    ht_in   = nondim_r * Cp* Tt_in
    
    # Compute the fuel to air ratio using turbine exit temperature, the fuel properties and freestream temperature
    f       = (ht4 - ht_in)/(eta_b*htf-ht4)

    # Computing the exit static and stagnation conditions
    ht_out  = Tt4 * Cp
    
    T_out     = Tt4/(1.+(gamma-1.)/2.*M0*M0)
    P_out     = Pt_out/((1.+(gamma-1.)/2.*M0*M0)**(gamma/(gamma-1.)))     
    
    # Pack results 
    combustor_conditions.outputs.stagnation_temperature  = Tt4
    combustor_conditions.outputs.stagnation_pressure     = Pt_out
    combustor_conditions.outputs.stagnation_enthalpy     = ht_out
    combustor_conditions.outputs.fuel_to_air_ratio       = f
    combustor_conditions.outputs.static_temperature      = T_out
    combustor_conditions.outputs.static_pressure         = P_out 
    combustor_conditions.outputs.mach_number             = M0 
    
    return 



# RCAIDE/Library/Methods/Powertrain/Converters/Compression_Nozzle/compute_compression_nozzle_performance.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------     

# package imports
import numpy as np  
from warnings import warn

# ---------------------------------------------------------------------------------------------------------------------- 
# compute_compression_nozzle_performance
# ----------------------------------------------------------------------------------------------------------------------    
def compute_compression_nozzle_performance(compression_nozzle, conditions):
    """
    Computes the performance of a compression nozzle based on its polytropic efficiency.
    
    Parameters
    ----------
    compression_nozzle : Data
        Data dictionary with compression nozzle properties
            - tag : str
                Identifier for the compression nozzle
            - pressure_ratio : float
                Pressure ratio across the nozzle [unitless]
            - polytropic_efficiency : float
                Polytropic efficiency of the nozzle [unitless]
            - pressure_recovery : float
                Pressure recovery factor [unitless]
            - compressibility_effects : bool
                Flag to include compressibility effects
            - working_fluid : Data
                Working fluid object with methods to compute properties
    conditions : Data
        Data dictionary with flow conditions
            - freestream : Data
                Freestream flow properties
                - pressure : ndarray
                    Freestream pressure [Pa]
                - mach_number : ndarray
                    Freestream Mach number [unitless]
            - energy : Data
                Energy conditions
                    - converters : dict
                        Dictionary of converter conditions indexed by tag
                        - inputs : Data
                            Input conditions
                            - stagnation_temperature : ndarray
                                Entering stagnation temperature [K]
                            - stagnation_pressure : ndarray
                                Entering stagnation pressure [Pa]
                            - static_temperature : ndarray
                                Entering static temperature [K]
                            - static_pressure : ndarray
                                Entering static pressure [Pa]
                            - mach_number : ndarray
                                Entering Mach number [unitless]
    
    Returns
    -------
    None
        Results are stored in conditions.energy.converters[compression_nozzle.tag].outputs:
            - stagnation_temperature : ndarray
                Exit stagnation temperature [K]
            - stagnation_pressure : ndarray
                Exit stagnation pressure [Pa]
            - stagnation_enthalpy : ndarray
                Exit stagnation enthalpy [J/kg]
            - mach_number : ndarray
                Exit Mach number [unitless]
            - static_temperature : ndarray
                Exit static temperature [K]
            - static_enthalpy : ndarray
                Exit static enthalpy [J/kg]
            - velocity : ndarray
                Exit nozzle velocity [m/s]
            - static_pressure : ndarray
                Exit static pressure [Pa]
    
    Notes
    -----
    This function computes the thermodynamic properties at the exit of a compression nozzle
    based on the inlet conditions and nozzle characteristics. It handles both subsonic and
    supersonic flows with appropriate relations.
    
    **Major Assumptions**
        * Pressure ratio and polytropic efficiency do not change with varying conditions
        * Adiabatic process
        * Subsonic or choked output
    
    **Theory**
    The compression nozzle performance is calculated using gas dynamics relations for
    compressible flow. For subsonic flow, isentropic relations are used. For supersonic flow,
    normal shock relations are applied. The stagnation properties are transformed to static
    properties based on the exit Mach number.
    
    References
    ----------
    [1] Stanford University, "AA283 Course Notes", https://web.stanford.edu/~cantwell/AA283_Course_Material/AA283_Course_Notes/
    
    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Converters.Expansion_Nozzle.compute_expansion_nozzle_performance
    """

    # Unpack conditions 
    P0                = conditions.freestream.pressure
    M0                = conditions.freestream.mach_number
    nozzle_conditions = conditions.energy.converters[compression_nozzle.tag]

    # Unpack inpust
    Tt_in                   = nozzle_conditions.inputs.stagnation_temperature
    Pt_in                   = nozzle_conditions.inputs.stagnation_pressure
    PR                      = compression_nozzle.pressure_ratio
    eta_p_old               = compression_nozzle.polytropic_efficiency
    eta_rec                 = compression_nozzle.pressure_recovery
    compressibility_effects = compression_nozzle.compressibility_effects
    
    # Unpack ram inputs
    working_fluid           = compression_nozzle.working_fluid
 
    # Compute the working fluid properties  
    T0     = nozzle_conditions.inputs.static_temperature
    P0     = nozzle_conditions.inputs.static_pressure   
    M0     = nozzle_conditions.inputs.mach_number 
    gamma  = working_fluid.compute_gamma(T0,P0) 
    Cp     = working_fluid.compute_cp(T0,P0)  
    a      = working_fluid.compute_speed_of_sound(T0,P0)    
 
    # Compute output stagnation quantities
    Pt_out  = Pt_in*PR*eta_rec
    Tt_out  = Tt_in*(PR*eta_rec)**((gamma-1)/(gamma*eta_p_old))
    ht_out  = Tt_out*Cp 

    if compressibility_effects: 
        # initilize arrays
        Pt_out   = np.ones_like(Tt_in)
        M_out     = np.ones_like(Tt_in)
        T_out    = np.ones_like(Tt_in)
        P_out    = np.ones_like(Tt_in)

        if M_out <= 1.0: # use isentropic relations
            i_low          = M0 <= 1.0
            Pt_out[i_low]  = Pt_in[i_low]*PR
            M_out[i_low]    = np.sqrt( (((Pt_out[i_low]/P0[i_low])**((gamma[i_low]-1.)/gamma[i_low]))-1.) *2./(gamma[i_low]-1.) ) 
            T_out[i_low]   = Tt_out[i_low]/(1.+(gamma[i_low]-1.)/2.*M_out[i_low]*M_out[i_low])
            P_out[i_low]   = Pt_out[i_low]/((1.+(gamma[i_low]-1.)/2.*M_out[i_low]*M_out[i_low])**(gamma[i_low]/(gamma[i_low]-1.)))

        else: # use normal shock
            i_high         = M0 > 1.0
            M_out[i_high]   = np.sqrt((1.+(gamma[i_high]-1.)/2.*M0[i_high]**2.)/(gamma[i_high]*M0[i_high]**2-(gamma[i_high]-1.)/2.))
            T_out[i_high]  = Tt_out[i_high]/(1.+(gamma[i_high]-1.)/2*M_out[i_high]*M_out[i_high])
            Pt_out[i_high] = PR*Pt_in[i_high]*((((gamma[i_high]+1.)*(M0[i_high]**2.))/((gamma[i_high]-1.)*\
                            M0[i_high]**2.+2.))**(gamma[i_high]/(gamma[i_high]-1.)))*((gamma[i_high]+1.)/(2.*gamma[i_high]*\
                            M0[i_high]**2.-(gamma[i_high]-1.)))**(1./(gamma[i_high]-1.))
            P_out[i_high]  = Pt_out[i_high]/(1.+(gamma[i_high]-1.)/2.*M_out[i_high]**2.)**(gamma[i_high]/(gamma[i_high]-1.))
    else:
        Pt_out  = Pt_in*PR*eta_rec 
        if np.any(Pt_out<P0): # in case pressures go too low
            warn('Pt_out goes too low',RuntimeWarning)
            Pt_out[Pt_out<P0] = P0[Pt_out<P0] 
        M_out   = np.sqrt( (((Pt_out/P0)**((gamma-1.)/gamma))-1.) *2./(gamma-1.) )
        T_out  = Tt_out/(1.+(gamma-1.)/2.*M_out*M_out)
        P_out  = Pt_out/(1.+(gamma-1.)/2.*M_out*M_out)**(gamma/(gamma-1.))
        
    # Compute exit ethalpy and velocity  
    h_out   = Cp*T_out
    u_out   = np.sqrt(2.*(ht_out-h_out))

    # Pack computed quantities into outputs
    nozzle_conditions.outputs.mach_number             = M_out
    nozzle_conditions.outputs.velocity                = u_out
    nozzle_conditions.outputs.static_enthalpy         = h_out
    nozzle_conditions.outputs.static_temperature      = T_out
    nozzle_conditions.outputs.static_pressure         = P_out
    nozzle_conditions.outputs.stagnation_enthalpy     = ht_out
    nozzle_conditions.outputs.stagnation_temperature  = Tt_out
    nozzle_conditions.outputs.stagnation_pressure     = Pt_out
    
    return 


# RCAIDE/Library/Plots/Performance/Aerodynamics/plot_airfoil_polars.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
from RCAIDE.Framework.Core import Units
from RCAIDE.Library.Plots.Common import set_axes, plot_style
import matplotlib.pyplot as plt 

# ----------------------------------------------------------------------------------------------------------------------
#  PLOTS
# ----------------------------------------------------------------------------------------------------------------------     

## @ingroup Library-Plots-Performance-Aerodynamics   
def plot_airfoil_polars(polar_data,
                        save_figure = False,
                        save_filename = "Airfoil_Polars",
                        file_type = ".png",
                        width = 11, height = 7):
    """
    Generate plots of airfoil aerodynamic characteristics from analysis results.

    Parameters
    ----------
    polar_data : Data
        Airfoil analysis results containing:
            - cl_invisc[0] : array
                Inviscid lift coefficients
            - cd_invisc[0] : array
                Inviscid drag coefficients
            - cm_invisc[0] : array
                Inviscid moment coefficients
            - AoA[0] : array
                Angles of attack [rad]
            - Re[0] : array
                Reynolds numbers

    save_figure : bool, optional
        Save figure to file if True, default False

    save_filename : str, optional
        Name for saved figure file, default "Airfoil_Polars"

    file_type : str, optional
        File extension for saved figure, default ".png"

    width : float, optional
        Figure width in inches, default 11

    height : float, optional
        Figure height in inches, default 7

    Returns
    -------
    fig : matplotlib.figure.Figure

    Notes
    -----
    Creates a 2x2 subplot figure showing inviscid airfoil characteristics:
        - Top left: Lift coefficient vs angle of attack
        - Top right: Drag coefficient vs angle of attack
        - Bottom left: Moment coefficient vs angle of attack
        - Bottom right: Lift-to-drag ratio vs angle of attack

    Results are plotted for a single Reynolds number case, with the
    Reynolds number shown in the legend (e.g., "Re=1.0e6").

    See Also
    --------
    RCAIDE.Library.Plots.Common.set_axes : Standardized axis formatting
    RCAIDE.Library.Plots.Common.plot_style : RCAIDE plot styling
    RCAIDE.Library.Analysis.Aerodynamics.compute_airfoil_inviscid : Analysis module
    RCAIDE.Library.Plots.Aerodynamics.plot_airfoil_polar_files : Viscous polar plotting function
    """ 
 
    # Get raw data polars 
    CL           = polar_data.cl_invisc[0]
    CD           = polar_data.cd_invisc[0]
    CM           = polar_data.cm_invisc[0]
    alpha        = polar_data.AoA[0]/Units.degrees
    Re_raw       = polar_data.Re[0]  
       
    Re_val = str(round(Re_raw[0])/1e6)+'e6' 
    
    # get plotting style 
    ps      = plot_style()  

    parameters = {'axes.labelsize': ps.axis_font_size,
                  'xtick.labelsize': ps.axis_font_size,
                  'ytick.labelsize': ps.axis_font_size,
                  'axes.titlesize': ps.title_font_size}
    plt.rcParams.update(parameters)
      
    fig   = plt.figure(save_filename)
    fig.set_size_inches(width,height) 
               
    axis_1 = plt.subplot(2,2,1)
    axis_1.plot(alpha, CL, color = ps.color, marker = ps.markers[0], linewidth = ps.line_width, label = 'Re = '+Re_val)
    axis_1.set_ylabel(r'$C_l$')
    set_axes(axis_1)    
    
    axis_2 = plt.subplot(2,2,2)
    axis_2.plot(alpha, CD, color = ps.color, marker = ps.markers[0], linewidth = ps.line_width)
    axis_2.set_ylabel(r'$C_d$')
    set_axes(axis_2) 

    axis_3 = plt.subplot(2,2,3)
    axis_3.plot(alpha, CM, color = ps.color, marker = ps.markers[0], linewidth = ps.line_width)
    axis_3.set_xlabel('AoA [deg]') 
    axis_3.set_ylabel(r'$C_m$')
    set_axes(axis_3) 
    
    axis_4 = plt.subplot(2,2,4)
    axis_4.plot(alpha, CL/CD, color = ps.color, marker = ps.markers[0], linewidth = ps.line_width)
    axis_4.set_xlabel('AoA [deg]')
    axis_4.set_ylabel(r'Cl/Cd')
    axis_4.set_ylim([-20,20])
    set_axes(axis_4) 
            
    if save_figure:
        plt.savefig(save_filename + file_type)   
    return  fig
     
     
     
     


# RCAIDE/Library/Methods/Weights/Correlation_Buildups/FLOPS/compute_horizontal_tail_weight.py
# 
# 
# Created:  Sep 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE 
from RCAIDE.Framework.Core    import Units
 
# ----------------------------------------------------------------------------------------------------------------------
#  Horizontal Tail Weight 
# ----------------------------------------------------------------------------------------------------------------------
def compute_horizontal_tail_weight(vehicle):
    """
    Calculates the weight of an aircraft's horizontal tail using FLOPS methodology.

    Parameters
    ----------
    vehicle : RCAIDE.Vehicle()
        Vehicle data structure containing:
            - wings.horizontal_tail.areas.reference
            - mass_properties.max_takeoff
            - flight_envelope.ultimate_load
            - design_dynamic_pressure
    

    Returns
    -------
    horizontal_tail_weight : float
        Estimated weight of the horizontal tail assembly in kilograms

    Notes
    -----
    The function implements the FLOPS (Flight Optimization System) weight estimation
    method for horizontal tail surfaces for general aviation aircraft. The calculation accounts 
    for geometric parameters and design loads based on maximum takeoff weight. 
    
    **Major Assumptions**
        * The weight is a function of the horizontal tail area, ultimate load factor, takeoff weight, and design dynamic pressure
    
    **Theory**

    The FLOPS horizontal tail weight estimation follows:

    .. math::
        W_{ht} = 0.016 * S_{ht}^{0.873} * (N_{ult} * W_{TO})^{0.414} * q^{0.122}

    Where:
        - W_{ht} is horizontal tail weight [lb]
        - S_{ht} is horizontal tail area [ft^2]
        - N_{ult} is ultimate load factor
        - W_{TO} is maximum takeoff weight [lb]
        - q is design dynamic pressure [psf]
    """
    SHT     = vehicle.wings.horizontal_stabilizer.areas.reference / Units.ft **2
    DG      = vehicle.mass_properties.max_takeoff / Units.lbs
    QCRUS   = vehicle.flight_envelope.design_dynamic_pressure / Units.psf
    ULF     = vehicle.flight_envelope.ultimate_load   
    WHT     = 0.016*SHT**0.873 * ( ULF * DG)**0.414 * QCRUS**0.122 
    return WHT * Units.lbs



# RCAIDE/Library/Missions/Common/Unknowns/ground.py
# 
# 
# Created:  Jul 2023, M. Clarke
 
# ----------------------------------------------------------------------------------------------------------------------
#  Unpack Unknowns
# ---------------------------------------------------------------------------------------------------------------------- 
def ground(segment):
    """
    Updates ground segment states from solver unknowns

    Parameters
    ----------
    segment : Segment
        The mission segment being analyzed
            - state:
                - unknowns:
                    - ground_velocity : array
                        Solved ground velocity [m/s]
                    - elapsed_time : float
                        Segment duration [s]
                - conditions.frames.inertial:
                    - velocity_vector : array
                        Vehicle velocity [m/s]
                    - time : array
                        Segment time points [s]
                - numerics.dimensionless:
                    - control_points : array
                        Normalized time points [-]
            - air_speed_start : float
                Initial airspeed [m/s]

    Returns
    -------
    None
        Updates segment conditions directly

    Notes
    -----
    This function applies ground-specific solver values to the segment state,
    handling velocity and time variables for ground operations. It ensures
    proper velocity transitions and time scaling.

    The function processes:
        1. Ground velocity application
        2. Time vector computation
        3. Initial velocity preservation

    **Major Assumptions**
        * Continuous velocity transitions
        * Valid time scaling
        * Non-negative velocities
        * Well-defined initial conditions

    See Also
    --------
    RCAIDE.Framework.Mission.Segments.Ground
    """       
    
    # unpack unknowns 
    ground_velocity = segment.state.unknowns.ground_velocity
    time            = segment.state.unknowns.elapsed_time
    
    # unpack givens
    v0         = segment.air_speed_start  
    t_initial  = segment.state.conditions.frames.inertial.time[0,0]
    t_nondim   = segment.state.numerics.dimensionless.control_points
    
    # time
    t_final    = t_initial + time  
    times      = t_nondim * (t_final-t_initial) + t_initial     

    # apply unknowns
    conditions = segment.state.conditions
    conditions.frames.inertial.velocity_vector[1:,0] = ground_velocity
    conditions.frames.inertial.velocity_vector[0,0]  = v0
    conditions.frames.inertial.time[:,0]             = times[:,0]



# RCAIDE/Library/Methods/Propulsor/Ducted_Fan_Propulsor/run_dfdc_analysis.py
# 
# Created: Sep 2024, M. Clarke 

# ---------------------------------------------------------------------------------------------------------------------- 
#  Imports
# ----------------------------------------------------------------------------------------------------------------------

from RCAIDE.Framework.Core  import redirect 
import sys 
import subprocess
import os 
from .purge_files  import purge_files 

# ---------------------------------------------------------------------------------------------------------------------- 
# Run DFDC Analysis
# ----------------------------------------------------------------------------------------------------------------------   
def run_dfdc_analysis(dfdc_object, print_output):
    """
    Executes the DFDC (Ducted Fan Design Code) executable with the specified input deck.
    
    Parameters
    ----------
    dfdc_object : DFDCAnalysis
        Analysis object containing the following attributes:
            - settings : Data
                Configuration settings
                    - new_regression_results : bool
                        Flag to use existing results instead of running analysis
                    - filenames : Data
                        File path information
                            - log_filename : str
                                Path to log file for stdout
                            - err_filename : str
                                Path to error file for stderr
                            - dfdc_bin_name : str
                                Path to DFDC executable
                            - case : str
                                Case name for DFDC
            - current_status : Data
                Current analysis state
                    - deck_file : str
                        Path to input deck file with DFDC commands
    print_output : bool
        Flag to control whether DFDC console output is displayed
    
    Returns
    -------
    exit_status : int
        Return code from the DFDC process
        0 indicates successful execution
        Non-zero values indicate errors
    
    Notes
    -----
    This function handles the execution of the DFDC executable by:
        1. Purging any existing log and error files
        2. Opening the input deck file and feeding commands to DFDC
        3. Capturing and redirecting stdout/stderr to specified files
        4. Returning the process exit code
    
    This function requires the third party DFDC executable to be installed. More
    information on the DFDC executable from Drela and Youngrencan be found at: https://web.mit.edu/drela/Public/web/dfdc/ 
    
    If new_regression_results is True, the function will skip execution
    and return a success code (0) without running DFDC.
    
    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Converters.Ducted_Fan.purge_files
    RCAIDE.Framework.Core.redirect
    """      
    new_regression_results = dfdc_object.settings.new_regression_results
    if new_regression_results:
        exit_status = 0 
    else:
        log_file = dfdc_object.settings.filenames.log_filename
        err_file = dfdc_object.settings.filenames.err_filename
        if isinstance(log_file,str):
            purge_files(log_file)
        if isinstance(err_file,str):
            purge_files(err_file)
        dfdc_call = dfdc_object.settings.filenames.dfdc_bin_name
        case      = dfdc_object.settings.filenames.case
        in_deck   = dfdc_object.current_status.deck_file  
    
        with redirect.output(log_file,err_file): 
            with open(in_deck,'r') as commands: 
                
                # Initialize suppression of console window output
                if print_output == False:
                    devnull = open(os.devnull,'w')
                    sys.stdout = devnull       
                    
                # Run DFDC
                dfdc_run = subprocess.Popen([dfdc_call,case],stdout=sys.stdout,stderr=sys.stderr,stdin=subprocess.PIPE)
                for line in commands:
                    dfdc_run.stdin.write(line.encode('utf-8'))
                    dfdc_run.stdin.flush()
                    
                  
                # Terminate suppression of console window output  
                if print_output == False:
                    sys.stdout = sys.__stdout__                    
                    
            dfdc_run.wait()
    
            exit_status = dfdc_run.returncode 

    return exit_status




# RCAIDE/Library/Methods/Powertrain/Converters/Motor/compute_motor_performance.py

# 
# Created:  Jan 2025, M. Clarke, M. Guidotti

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE imports
import  RCAIDE
from RCAIDE.Framework.Core import  Units

# package imports 
import numpy as np
 
# ----------------------------------------------------------------------------------------------------------------------
#  compute_omega_and_Q_from_Cp_and_V
# ----------------------------------------------------------------------------------------------------------------------    
def compute_motor_performance(motor,conditions):
    """
    Computes motor performance characteristics including electrical, mechanical and thermal parameters.

    Parameters
    ----------
    motor : Converter
        Motor component (DC_Motor or PMSM_Motor) for which performance is being computed
    motor_conditions : Conditions
        Container for motor operating conditions
    conditions : Conditions 
        Mission segment conditions containing freestream properties

    Returns
    -------
    None
        Updates motor_conditions in-place with computed performance parameters

    Notes
    -----
    This function handles both PMSM and DC motor types with different computation approaches: 

    For PMSM motors:
        - Computes electromagnetic torque and power
        - Accounts for gearbox effects
        - Calculates thermal resistances and heat transfer
        - Determines cooling flow characteristics
        - Evaluates airgap and endspace heat transfer

    For DC motors:
        - Uses speed-torque relationships
        - Accounts for gearbox effects 
        - Determines overall efficiency
        
    For Both Motors:
        - motor.inverse_calculation arg is used to determine parameters that are solved.
        - motor.inverse_calculation == False calculates electrical properties (electrical power and voltage) from mechnical properties (rpm and torque)
        - motor.inverse_calculation == True calculates mechanial properties (rpm and torque) from electrical properties (electrical power and voltage)

    **Major Assumptions**
        * Steady state operation
        * Uniform temperature distribution
        * No magnetic saturation effects
        * Linear speed-torque characteristics for DC motors
        * Constant material properties

    See Also
    --------
    RCAIDE.Library.Components.Powertrain.Converters.DC_Motor
    RCAIDE.Library.Components.Powertrain.Converters.PMSM_Motor
    """           
    # Unpack
    motor_conditions = conditions.energy.converters[motor.tag]
    
    if (type(motor) == RCAIDE.Library.Components.Powertrain.Converters.PMSM_Motor): 
        if motor.inverse_calculation == False:
            Res            = motor.resistance
            G              = motor.gearbox.gear_ratio
            I              = motor_conditions.inputs.current
            V              = motor_conditions.inputs.voltage
            I_turn         = I/motor.number_of_turns                                                             # [A]            current in each turn
            omega          = (motor.speed_constant*(V - I*Res)*Units.rpm) /G                                     # [RPM -> rad/s] rotor angular velocity
            A              = np.pi * ((motor.stator_outer_diameter**2 - motor.stator_inner_diameter**2) / 4)     # [m**2]         cross-sectional area of the reluctance path perpendicular to length 𝑙    
            MMF_coil       = motor.number_of_turns*I_turn                                                        # [A*turns]      magnetomotive force applied to the reluctance path for a coil (Eq.14)  
            R              = motor.length_of_path/(A*motor.mu_0*motor.mu_r)                                      # [A*turn/Wb]    reluctance of a given path or given reluctant element (Eq.16) 
            phi            = MMF_coil/R                                                                          # [Wb]           magnetic flux through the reluctance path (Eq.12)
            B_sign         = phi/A                                                                               # [V*s/m**2]     ranges from 0.5 to 1.2, average magnitude of the radial flux density produced by the rotor
            A_sign         = (motor.winding_factor*I)/(np.pi*motor.stator_inner_diameter)                        # [-]            stator electrical loading (Eq.2)        
            TQ             = (np.pi/2)*(B_sign*A_sign)*(motor.stator_inner_diameter**2)*motor.motor_stack_length # [Nm]           torque (Eq.1)
            P              = omega*TQ                                                                            # [W]            power (Eq.1)        
            A              = np.pi * ((motor.stator_outer_diameter**2 - motor.stator_inner_diameter**2) / 4)     # [m**2]         cross-sectional area of the reluctance path perpendicular to length 𝑙    
            R_cond_path    = motor.length_of_conductive_path/(motor.thermal_conductivity*A)                      # [K/W]          Conductive Path Thermal Resistance (Eq.68)
    
            if motor.Conduction_laminar_flow == True:   
                Nu     = 0.453*(motor.Re_cooling_flow**0.5)*(motor.Prandtl_number**(1/3))                               # Laminar Nusselt number (Eq.71)
            else:
                Nu   = 0.0308*(motor.Re_cooling_flow**(4/5))*(motor.Prandtl_number**(1/3))                              # Turbulent Nusselt number (Eq.71) 
            h    = motor.characteristic_length_of_flow*Nu*motor.thermal_conductivity_fluid               # [W/m**2*K]     convection coefficient of the flow at a liquid to solid interfaced
            R_conv_path    = 1/(h*A)                                                                       # [K/W]          Fluid Flow Thermal Resistance (Eq.69)
    
            if motor.Re_cooling_flow < 3000:
                Nu_cooling_flow     = 1.051*np.log(motor.height_of_duct/motor.width_of_duct) + 2.89         # Nusselt number for cooling flow in rectangular ducts and Re_d < 3000 (Eq.72)
                h_cooling_flow      = motor.characteristic_length_of_flow*Nu_cooling_flow*motor.thermal_conductivity_fluid                   # [W/m**2*K]     convection coefficient of the flow at a liquid to solid interfaced
                R_conv_path_cooling_flow    = 1/(h_cooling_flow*A)                             # [K/W]          Fluid Flow Thermal Resistance (Eq.69)
            else:
                if motor.Convection_laminar_flow == True:
                    f = 64/motor.Re_cooling_flow
                else:
                    f = (0.79*np.log(motor.Re_cooling_flow) - 1.64)**(-2)                                   # Turbulent Moody friction factor (Eq.73)
                Nu_cooling_flow   = ((f/8)*(motor.Re_cooling_flow - 1000)*motor.Prandtl_number)/(1 + 12.7*((f/8)**0.5)*(motor.Prandtl_number**(2/3) - 1)) # Nusselt number for cooling flow in rectangular ducts and Re_d >= 3000 (Eq.72)
                h_cooling_flow    = motor.characteristic_length_of_flow*Nu_cooling_flow*motor.thermal_conductivity_fluid                     # [W/m**2*K]     convection coefficient of the flow at a liquid to solid interfaced
                R_conv_path_cooling_flow    = 1/(h_cooling_flow*A)                                                                             # [K/W]          Fluid Flow Thermal Resistance (Eq.69)
    
            Delta_P_flow               = ((f*motor.density_of_fluid*motor.velocity_of_fluid**2)/(2*motor.hydraulic_diameter_of_duct))*motor.length_of_channel # Flow pressure drop (Eq.74)
            Loss_cooling               = Delta_P_flow*motor.volume_flow_rate_of_fluid                              # Power needed to pump the fluid through the duct (Eq.75)
    
            if motor.Taylor_number < 41: 
                Nu_airgap  = 2                                                                 # Nusselt number for the airgap convection and Ta < 41 (Eq.76)
            elif motor.Taylor_number> 41 and motor.Taylor_number < 100:
                Nu_airgap = 0.202*(motor.Taylor_number**(0.63))*(motor.Prandtl_number**0.27)                          # Nusselt number for the airgap convection and 41 < Ta < 100 (Eq.76)
            else:
                Nu_airgap = 0.386*(motor.Taylor_number**0.5)*(motor.Prandtl_number**0.27)                             # Nusselt number for the airgap convection and 100 < Ta (Eq.76)
            h_airgap    = motor.characteristic_length_of_flow*Nu_airgap*motor.thermal_conductivity_fluid                                     # [W/m**2*K]     convection coefficient of the flow at a liquid to solid interfaced
            R_airgap    = 1/(h_airgap*A)                                                       # [K/W]          Fluid Flow Thermal Resistance (Eq.69)
    
            if motor.axial_gap_to_radius_of_rotor == 0.01:
                if motor.Re_airgap < 1e5:
                    Nu_G          = 7.46*motor.Re_airgap**(0.32)                               # Nusselt number for laminar flow and G = 0.01 (Eq.77)
                else:
                    Nu_G        = 0.044*motor.Re_airgap**(0.75)                                # Nusselt number for turbulent flow and G = 0.01 (Eq.78)  
            elif motor.axial_gap_to_radius_of_rotor> 0.02 and motor.axial_gap_to_radius_of_rotor < 0.06:
                if motor.Re_airgap < 1e5:
                    Nu_G     = 0.5*(1 + 5.47*(10**-4)*np.exp(112*motor.axial_gap_to_radius_of_rotor))*(motor.Re_airgap**0.5) # Nusselt number for laminar flow and G = 0.02 - 0.06 (Eq.77)
                else:    
                    Nu_G  = 0.5*(12.57*np.exp(-33.18*motor.axial_gap_to_radius_of_rotor))*(motor.Re_airgap**(0.6 + 25*motor.axial_gap_to_radius_of_rotor**(12/7))) # Nusselt number for turbulent flow and G = 0.02 - 0.06 (Eq.78)    
            elif motor.axial_gap_to_radius_of_rotor > 0.06:
                if motor.Re_airgap < 1e5:
                    Nu_G = 0.35*(motor.Re_airgap**0.5)                                         # Nusselt number for laminar flow and G > 0.06 (Eq.77)
                else:    
                    Nu_G = 0.0151*(motor.Re_airgap**0.6)                                       # Nusselt number for turbulent flow and G > 0.06 (Eq.78)
            h_endspace    = motor.characteristic_length_of_flow*Nu_G*motor.thermal_conductivity_fluid                                        # [W/m**2*K]     convection coefficient of the flow at a liquid to solid interfaced
            R_endspace    = 1/(h_endspace*A)                                                   # [K/W]          Fluid Flow Thermal Resistance (Eq.69)
    
            Q_cond_path              = motor.Delta_T/R_cond_path                               # heat through a conductive thermal path (Eq.67)
            Q_conv_path              = motor.Delta_T/R_conv_path                               # heat through a thermal path (Eq.67)
            Q_conv_path_cooling_flow = motor.Delta_T/R_conv_path_cooling_flow                  # heat through a thermal path (Eq.67)
            Q_conv_airgap            = motor.Delta_T/R_airgap                                  # heat through a thermal path (Eq.67)
            Q_conv_endspace          = motor.Delta_T/R_endspace                                # heat through a thermal path (Eq.67)
    
            omega_gearbox  = omega * G
            TQ_gearbox     = TQ / G
            
            motor_conditions.Q_cond_path              = Q_cond_path                            
            motor_conditions.Q_conv_path              = Q_conv_path                            
            motor_conditions.Q_conv_path_cooling_flow = Q_conv_path_cooling_flow               
            motor_conditions.Q_conv_airgap            = Q_conv_airgap                          
            motor_conditions.Q_conv_endspace          = Q_conv_endspace                        
            motor_conditions.Loss_cooling             = Loss_cooling                           
            motor_conditions.outputs.torque           = TQ_gearbox 
            motor_conditions.outputs.omega            = omega_gearbox 
            motor_conditions.outputs.power            = P
        else:
            io              = motor.no_load_current
            G               = motor.gearbox.gear_ratio 
            omega_gearbox   = motor_conditions.outputs.omega 
            omega           = omega_gearbox / G
            power           = motor_conditions.outputs.power  
            Kv              = motor.speed_constant 
            D_in            = motor.stator_inner_diameter  
            kw              = motor.winding_factor  
            Res             = motor.resistance                      
            L               = motor.motor_stack_length                    
            l               = motor.length_of_path                  
            mu_0            = motor.mu_0                            
            mu_r            = motor.mu_r   
            Q               = power/omega                                      
            i               = np.sqrt((2*(Q)*l)/(D_in*mu_0*mu_r*L*kw))    
            v               = omega/Kv   + ((Q*Kv) + io) * Res  
            etam            = (1-io/i)*(1-i*Res/v)     

            motor_conditions.inputs.power    = v * i   
            motor_conditions.inputs.voltage  = v    
            motor_conditions.inputs.current  = i 
            motor_conditions.efficiency      = etam             
            
    else:
        if motor.inverse_calculation == False:
            G              = motor.gearbox.gear_ratio  
            Res            = motor.resistance  
            Kv             = motor.speed_constant
            G              = motor.gearbox.gear_ratio 
            io             = motor.no_load_current  
            v              = motor_conditions.inputs.voltage
             
            i              = motor_conditions.inputs.current 
            omega          = ((v - (Res * i)) * Kv)  
            Q              = ((v- (omega) /Kv)/Res -io)/Kv  
            etam           = (1-io/i)*(1-i*Res/v)
            omega_gearbox  = omega * G
            Q_gearbox      = Q / G
                
            motor_conditions.outputs.torque = Q_gearbox
            motor_conditions.outputs.omega  = omega_gearbox 
            motor_conditions.outputs.power  = omega_gearbox*Q_gearbox 
            motor_conditions.efficiency     = etam
            
        else: 
            G              = motor.gearbox.gear_ratio 
            omega_gearbox  = motor_conditions.outputs.omega 
            power          = motor_conditions.outputs.power  
            io             = motor.no_load_current  
            Q_gearbox      = power / omega_gearbox   
            omega          = omega_gearbox / G
            Q              = Q_gearbox * G
            Res            = motor.resistance  
            Kv             = motor.speed_constant  
            io             = motor.no_load_current
            v              = ((Q *Kv) + io) * Res +  omega/Kv 
            i              = (v - (omega)/Kv)/Res  
            P              = i * v 
            etam           = (1-io/i)*(1-i*Res/v)

            motor_conditions.inputs.power    = v * i   
            motor_conditions.inputs.voltage  = v    
            motor_conditions.inputs.current  = i 
            motor_conditions.efficiency      = etam 
                   
    return


#  RCAIDE/Methods/Energy/Distributors/Electrical_Bus/initialize_bus_properties.py
# 
# Created:  Sep 2024, S. Shekar
# Modified: Jan 2025, M. Clarke
#
from RCAIDE.Library.Methods.Powertrain.Sources.Batteries.Common          import compute_module_properties 
from RCAIDE.Library.Methods.Powertrain.Converters.Fuel_Cells.Common             import compute_stack_properties

# ----------------------------------------------------------------------------------------------------------------------
#  METHODS
# ---------------------------------------------------------------------------------------------------------------------- 
def initialize_bus_properties(bus): 
    """
    Initializes the bus electrical properties based on what is appended onto the bus.
    
    Parameters
    ----------
    bus : ElectricalBus
        The electrical bus component with the following attributes:
            - battery_modules : list
                List of battery modules connected to the bus
            - battery_module_electric_configuration : str
                Configuration of battery modules ('Series' or 'Parallel')
            - fuel_cell_stacks : list
                List of fuel cell stacks connected to the bus
            - fuel_cell_stack_electric_configuration : str
                Configuration of fuel cell stacks ('Series' or 'Parallel')
    
    Returns
    -------
    None
        This function modifies the bus object in-place, setting the following attributes:
            - voltage : float
                Bus voltage [V]
            - nominal_capacity : float
                Nominal capacity [Ah]
            - maximum_energy : float
                Maximum energy storage capacity [J]
    
    Notes
    -----
    This function calculates the electrical properties of the bus based on the connected
    energy sources (battery modules and fuel cell stacks). It handles both series and 
    parallel configurations.
    
    For battery modules:
        - In series configuration: voltages add, capacity is the maximum of all modules
        - In parallel configuration: voltage is the maximum voltage of all modules, capacities add
    
    For fuel cell stacks:
        - In series configuration: voltages add
        - In parallel configuration: voltage is the maximum voltage of all stacks
    
    The function first computes properties for each individual module/stack by calling
    their respective property computation functions.
    
    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Sources.Batteries.Common.compute_module_properties
    RCAIDE.Library.Methods.Powertrain.Converters.Fuel_Cells.Common.compute_stack_properties
    """
    if len(bus.battery_modules) > 0: 
        if bus.battery_module_electric_configuration == 'Series':
            bus.nominal_capacity = 0
            bus.maximum_energy   = 0
            for battery_module in  bus.battery_modules: 
                compute_module_properties(battery_module) 
                bus.voltage         +=  battery_module.voltage
                bus.maximum_energy  +=  battery_module.maximum_energy
                bus.nominal_capacity =  max(battery_module.nominal_capacity, bus.nominal_capacity)  
        elif bus.battery_module_electric_configuration == 'Parallel':
            bus.voltage = 0
            bus.maximum_energy   = 0
            for battery_module in  bus.battery_modules: 
                compute_module_properties(battery_module)        
                bus.voltage           =  max(battery_module.voltage, bus.voltage)
                bus.nominal_capacity +=  battery_module.nominal_capacity        
                bus.maximum_energy  +=  battery_module.initial_maximum_energy
    
    cumulative_fuel_cell_stack =  0
    if len(bus.fuel_cell_stacks) > 0: 
        if bus.fuel_cell_stack_electric_configuration == 'Series':
            bus.maximum_energy   = 0
            for fuel_cell_stack in  bus.fuel_cell_stacks: 
                compute_stack_properties(fuel_cell_stack)
                cumulative_fuel_cell_stack += fuel_cell_stack.voltage 
            bus.voltage  =  min(fuel_cell_stack.voltage, cumulative_fuel_cell_stack) 
        elif bus.fuel_cell_stack_electric_configuration == 'Parallel': 
            for fuel_cell_stack in  bus.fuel_cell_stacks: 
                compute_stack_properties(fuel_cell_stack)        
                bus.voltage     =  max(fuel_cell_stack.voltage, bus.voltage)              
    return